import os

from .settings import *
from .rst_parser import ParseRST
from .section_parameters import ParametersSection
from .section_references import ReferencesSection
from .section_settings import SettingsSection
from ..documentation import Documentation

class ExportDocsRST:
        
    docs: Documentation

    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass
    
    def __init__(self, docs: Documentation, **kwargs):
        self.docs = docs
        
    def build(self, dir_build: str, title: str, intro=None):

        sections = {
            "Parameters": {
                'filename':'parameters',
                'content':  ParametersSection,
                'sections': {
                    "types":"Node types",
                    "parameters":"Parameter list",
                    "nodes":"Parameter nodes"
                }
            },
            "References": {
                'filename':'references',
                'content':  ReferencesSection,
                'sections': {
                    "injections":"Injections",
                    "imports":"Imports"
                }
            },
            "Settings":   {
                'filename':'settings',
                'content':  SettingsSection,
                'sections': {
                    "units":"Units",
                    "sources":"Sources"
                }
            },
        }        
        
        # parse index page

        rst = ParseRST()
        rst.title(0, title)
        
        if intro:
            rst.paragraph(intro)
        
        pages = [values['filename'] for key, values in sections.items()]
        rst.toc(pages, maxdepth=2, caption="Contents")
        
        with open(f"{dir_build}/index.rst", "w") as file:
            file.write(rst.parse())

        for name, settings in sections.items():
            with settings['content'](self.docs) as section:
                section.rst.title(0, name)
                with open(f"{dir_build}/{settings['filename']}.rst", "w") as file:
                    file.write(section.build())



import re
import numpy as np

from . import Units
from .matter import Matter
from .composite import Component
from .periodic_table import *
from ..units import Quantity, Unit
from .. import RowCollector, ParameterTable

PERIODIC_TABLE = ParameterTable(PT_HEADER, PT_DATA, keys=True)

class Element(Component, Matter):
    natural: bool
    element: str
    isotope: int
    ionisation: int
    Z: float = None
    N: float = None
    e: float = None

    # get a specific isotope
    def get_isotope(self, element:str, isotope: int, ionisation: int):
        isotopes = PERIODIC_TABLE[element]
        Z = isotopes.Z
        iso = isotope if isotope else Z*2
        ion = ionisation if ionisation else 0
        if str(iso) not in isotopes.A:
            raise Exception("Required isotope could not be found:", element, iso)
        M, NA = isotopes.A[str(iso)]
        A = Quantity(M, Units.ATOMIC_MASS) + Quantity(ion, '[m_e]')
        N = int(iso-isotopes.Z)
        e = isotopes.Z+ion
        return NA, A, Z, N, e, iso, ion
        
    # get the most abundant isotope
    def get_abundant(self, element:str, ionisation:int):
        isotopes = PERIODIC_TABLE[element]
        idmax = np.argmax([iso[1] for iso in isotopes.A.values()])
        isotope = int(list(A for A in isotopes.A.keys())[idmax])
        return self.get_isotope(element, isotope, ionisation)        

    # get a natural isotopic average
    def get_natural(self, element:str, ionisation:int):
        isotopes = PERIODIC_TABLE[element]
        with RowCollector(['NA', 'A', 'Z', 'N', 'e', 'iso', 'ion']) as rc:
            for iso in isotopes.A.keys():
                rc.append(self.get_isotope(element, int(iso), ionisation))
            return (
                np.sum(rc.NA), 
                np.average(rc.A, weights=rc.NA),
                np.average(rc.Z, weights=rc.NA),
                np.average(rc.N, weights=rc.NA),
                np.average(rc.e, weights=rc.NA),
                np.average(rc.iso, weights=rc.NA),
                int(rc.ion[0]),
            )

    def __init__(self, expr:str, proportion:int=1, natural:bool=True, **kwargs):
        Matter.__init__(self, **kwargs)
        Component.__init__(self, proportion)
        self.expr = expr
        self.natural = natural
        # parse the expr
        if m := re.match("(\[(p|n|e)\])", expr):
            nucleon = m.group(2)
            self.element = expr
            NA, iso, ion = 100.0, None, None
            self.mass = Unit(f"[m_{nucleon}]").to(Units.ATOMIC_MASS)
            nucleons = {'p':(1,0,0),'n':(0,1,0),'e':(0,0,1)}
            self.Z, self.N, self.e = nucleons[nucleon]
            self.isotope = 0
            self.ionisation = 0
        elif m := re.match("([a-zA-Z]{1,2})(\{([0-9]+)([+-]{1}[0-9]*)\}|\{([0-9]+)\}|\{([+-]{1}[0-9]*)\}|)", expr):
            # Extract information about isos
            element, variant, iso1, ion1, iso2, ion3 = m.groups()
            if element=='D':
                element, variant, iso1, ion1, iso2, ion3 = 'H', True, 2, ion1, 2, None
            elif element=='T':
                element, variant, iso1, ion1, iso2, ion3 = 'H', True, 3, ion1, 3, None
            self.element = element
            if iso1 and ion1:
                if ion1=="-": ion1="-1"
                elif ion1=="+": ion1="1"
                self.isotope, self.ionisation = int(iso1), int(ion1)
            elif iso2:
                self.isotope, self.ionisation = int(iso2), 0
            elif ion3:
                if ion3=="-": ion3="-1"
                elif ion3=="+": ion3="1"
                self.isotope, self.ionisation = None, int(ion3)
            else:
                self.isotope, self.ionisation = None, 0
            # set element values
            if self.isotope:
                NA, self.mass, self.Z, self.N, self.e, iso, ion = self.get_isotope(self.element, self.isotope, self.ionisation)
            elif self.natural:
                NA, self.mass, self.Z, self.N, self.e, self.isotope, self.ionisation = self.get_natural(self.element, self.ionisation)
            else:
                NA, self.mass, self.Z, self.N, self.e, self.isotope, self.ionisation = self.get_abundant(self.element, self.ionisation)
        else:
            raise Exception('Unrecognized expr', expr)
        self.component_mass = self.mass
        self.composite_mass = self.mass
        Matter._norm(self)

    def __mul__(self, other:float):
        return Element(self.expr, self.proportion*other, natural=self.natural)
    
    def __add__(self, other:'Element'):
        if self.expr!=other.expr:
            raise Exception("Only same elements can be added up:", self.expr, other.expr)
        return Element(self.expr, self.proportion+other.proportion, natural=self.natural)
        
    def __str__(self):
        if self.proportion>1:
            return f"Element({self.expr}{self.proportion} mass={self.proportion*self.component_mass.value(Units.ATOMIC_MASS):.3f} Z={self.proportion*self.Z} N={self.proportion*self.N:.3f} e={self.proportion*self.e})"
        else:
            return f"Element({self.expr} mass={self.component_mass.value(Units.ATOMIC_MASS):.3f} Z={self.Z} N={self.N:.3f} e={self.e})"
    
    def _print_table(self, columns:dict, fn_data:callable, **kwargs):
        df = fn_data(quantity=False, **kwargs).to_dataframe()
        df = df.rename(columns={k:f"{k}[{v}]" for k,v in columns.items() if v})
        print( df.to_string(index=False) )

    def _data(self, columns:dict, fn_row:callable, stats:bool=False, weight:bool=False, components:list=None, quantity:bool=True):
        column_names = list(columns.keys())
        pt = ParameterTable(column_names, keys=True, keyname='expr')
        values = fn_row(self.expr, self)
        row = []
        for col in column_names:
            value, unit = values[col], columns[col]
            if quantity and unit: # returning quantities
                if isinstance(value, Quantity):
                    row.append(value.to(unit))
                else:
                    row.append(Quantity(value, unit))
            elif isinstance(value, Quantity): # returning scalar
                    row.append(value.value(unit) if unit else value.value())
            else: # returning scalar
                row.append(value)
        pt[self.expr] = row
        return pt
        
    def print(self):
        text =  "Element:\n\n"
        if self.proportion>1:
            text += f"Expression: {self.expr}{self.proportion}\n"
            text += f"Mass:       {self.proportion*self.component_mass.value(Units.ATOMIC_MASS):.3f}\n"
            text += f"Protons:    {self.proportion*self.Z}\n"
            text += f"Neutrons:   {self.proportion*self.N:.3f}\n" 
            text += f"Electrons:  {self.proportion*self.e}\n"
        else:
            text += f"Expression: {self.expr}\n"
            text += f"Mass:       {self.component_mass.value(Units.ATOMIC_MASS):.3f}\n"
            text += f"Protons:    {self.Z}\n"
            text += f"Neutrons:   {self.N:.3f}\n"
            text += f"Electrons:  {self.e}"
        print(text)
        if self.mass_density:
            print("")
            Matter._print(self)


from PIL import Image
import numpy as np
import matplotlib.pyplot as plt
from typing import Union

class ThumbnailImage:
    """ Create a thumbnail from an image

    :param data: Image data array or a file name string
    :param tuple extent: Extent of an image (xmin, xmax, ymin, ymax)
    :param str mode: Image mode F/RGB/RGBA
    """
    
    extent: list
    im: Image = None
    
    def __init__(self, data, extent:list=None, mode=None):
        if isinstance(data, str):
            if mode is None: mode = 'RGB'
            self.im = Image.open(data).convert(mode)
        else:
            if mode is None: mode = 'F'
            self.im = Image.fromarray(data).convert(mode)
        self.extent = list(extent if extent else [0,1,0,self.im.size[1]/self.im.size[0]])
        self.xratio = np.abs(self.im.size[0]/(self.extent[1]-self.extent[0]))
        self.yratio = np.abs(self.im.size[1]/(self.extent[3]-self.extent[2]))
           
    def crop(self, left:Union[float,tuple], right:float=None, bottom:float=None, top:float=None, bgcolor=0):
        """ Change image extent

        :param left: Tuple with an image extent, or left extent
        :param float right: Right extent
        :param float bottom: Bottom extent
        :param float top: Top extent
        :param bgcolor: Color of the thumbnail padding (float or a tuple)
        """
        if isinstance(left, (list,tuple)) and right is None and bottom is None and top is None:
            extent = left  
        elif left is None and right is None and bottom is None and top is None:
            raise Exception("This function requires at least one extent:", left, right, botom, top)
        else:
            extent = list(self.extent)
            if left is not None:   extent[0] = left
            if right is not None:  extent[1] = right
            if bottom is not None: extent[2] = bottom
            if top is not None:    extent[3] = top
        xpix = int(np.round((extent[1]-extent[0])*self.xratio))
        ypix = int(np.round((extent[3]-extent[2])*self.yratio))
        layer = ThumbnailImage(
            extent = extent,
            data = np.full((ypix,xpix),bgcolor).astype(np.asarray(self.im).dtype),
            mode = self.im.mode,
        )
        pos = [int((layer.im.size[i]-self.im.size[i])/2) for i in range(2)]
        layer.im.paste(self.im, pos)
        self.im = layer.im
        self.extent = list(extent)
        self.xratio = np.abs(self.im.size[0]/(extent[1]-extent[0]))
        self.yratio = np.abs(self.im.size[1]/(extent[3]-extent[2]))
        return self

    def resize(self, xres:int = None, yres:int = None):
        """ Change image resolution
        
        :param int xres: Resolution on the X-axis
        :param int yres: Resolution on the Y-axis
        """
        if isinstance(xres, tuple):
            resolution = xres
        elif xres is None and yres is None:
            raise Exception("At least one resolution has to be set:", xres, yres)
        else:
            resolution = list(self.im.size)
            if xres is not None: resolution[0] = xres
            if yres is not None: resolution[1] = yres
        self.im = self.im.resize(resolution)
        return self
        
    def draw(self, ax=None, **kwargs):
        """ Draw current image on an axis

        :param ax: Matplotlib axis
        """
        if not ax: ax=plt
        return ax.imshow(np.asarray(self.im), extent=self.extent, origin='lower', **kwargs)
        
    def save(self, file_name:str, format:str=None):
        """ Save thumbnail as a file
        
        :param str file_name: Name of the image file
        """
        if self.im.mode != 'RGB':
            self.im = self.im.convert('RGB')
        self.im.save(file_name, format=format)



import numpy as np
from typing import Union, Dict

from . import Norm, Units
from .matter import Matter
from .. import ParameterTable, RowCollector
from ..units import Quantity

class Component:
    expr: str                       # expression
    proportion: Union[float,int]    # proportion 
    component_mass: Quantity        # mass of a component

    def __init__(self, proportion:int=1):
        self.proportion = proportion

class Composite:
    components: Dict[str,Component] # list of composite components
    component_class: callable       # name of the class that inherits Composite
    
    cols_components: dict           # settings of component columns
    cols_composit: dict             # settings of composite columns
    
    natural: bool                   # natural ements
    norm_type: Norm                 # type of composite norm
    proportion_norm: Union[float,int] # sum of all comonent proportions
    composite_mass: Quantity         # sum of all component masses
    
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass
    
    def __init__(self, 
        solver:callable, component_class:callable, expr:Union[str,dict], 
        cols_components: dict, cols_composite: dict,
        natural:bool, norm_type:Norm,
    ):
        self.expr = ''
        self.norm_type = norm_type
        self.natural = natural
        self.cols_components = cols_components
        self.cols_composite = cols_composite | {
            "x":     Units.FRACTION, 
            "X":     Units.FRACTION,
        }
        self.component_class = component_class
        self.components = {}
        if isinstance(expr,str) and expr:
            self.expr = expr
            with solver(self.atom) as ms:
                for expr, component in ms.solve(expr).components.items():
                    self.components[expr] = component
        elif isinstance(expr, dict) and expr:
            for expr, frac in expr.items():
                self.add(expr, frac)
        self._norm()
    
    def _norm(self):
        # calculate total coutn and mass of all components
        components = self.components.values()
        if self.norm_type==Norm.MASS_FRACTION:
            self.proportion_norm = np.sum([i.proportion/i.component_mass for i in components])
            self.composite_mass  = np.sum([i.proportion for i in components])
        else:
            self.proportion_norm = np.sum([i.proportion for i in components])
            self.composite_mass  = np.sum([i.proportion*i.component_mass for i in components])
        if type(self) in Component.__subclasses__():
            self.component_mass = self.composite_mass
        Matter._norm(self)

    def _data(self, columns:dict, fn_row:callable, stats:bool=False, weight:bool=False, components:list=None, quantity:bool=True):
        if not self.components: 
            return None   # no components
            
        # populate columns with values
        weights = []
        column_names = list(columns.keys())
        pt = ParameterTable(column_names, keys=True, keyname='expr')
        rc = RowCollector(column_names)
        for s,m in self.components.items():
            # component is not selected
            if components and s not in components:
                continue  
            # calculate matter properties
            row, values = [], fn_row(s,m)
            if self.norm_type in [Norm.NUMBER_FRACTION, Norm.NUMBER]:
                values['x'] = Quantity(m.proportion/self.proportion_norm)
                values['X'] = m.proportion*m.component_mass/self.composite_mass
                weights.append(m.proportion)
            elif self.norm_type==Norm.MASS_FRACTION:
                values['x'] = m.proportion/m.component_mass/self.proportion_norm
                values['X'] = Quantity(m.proportion/self.composite_mass)
                weights.append(m.proportion/m.component_mass)
            # convert to proper units
            for col in column_names:
                value, unit = values[col], columns[col]
                if quantity and unit: # returning quantities
                    if isinstance(value, Quantity):
                        row.append(value.to(unit))
                    else:
                        row.append(Quantity(value, unit))
                elif isinstance(value, Quantity): # returning scalar
                        row.append(value.value(unit) if unit else value.value())
                else: # returning scalar
                    row.append(value)
            # insert row
            pt[s] = row
            rc.append(row)
        
        # add rows with averages and sums
        if stats:
            if weight: # use weighted average
                if self.norm_type == Norm.NUMBER:
                    pt['avg'] = [np.average(np.divide(rc[p],weights), weights=weights) for p in column_names]
                else:
                    pt['avg'] = [np.average(rc[p]) for p in column_names]
            else:
                pt['avg'] = [np.average(rc[p]) for p in column_names]
            pt['sum'] = [np.sum(rc[p]) for p in column_names]
            
        return pt

    def _multiply(self, composite, other):
        for expr, component in self.components.items():
            composite.add(expr, component.proportion*other)
        return composite
    
    def _add(self, composite, other):
        for expr, component in self.components.items():
            composite.add(expr, component.proportion)
        if isinstance(other, self.component_class):
            composite.add(other.expr, other.proportion)
        else:
            for expr, component in other.components.items():
                composite.add(expr, component.proportion)
        return composite
    
    def _add_expr(self, expr:str, proportion:int):
        pass
    
    def add(self, expr:str, proportion:int=1):
        self._add_expr(expr, proportion)
        if expr in self.components:
            self.components[expr].proportion += proportion
        else:
            self.components[expr] = self.component_class(expr, natural=self.natural, proportion=proportion)
        self._norm()

    def _print_table(self, columns:dict, fn_data:callable, **kwargs):
        df = fn_data(quantity=False, **kwargs).to_dataframe()
        df = df.rename(columns={k:f"{k}[{v}]" for k,v in columns.items() if v})
        print( df.to_string(index=False) )
        
    def print_components(self):
        self._print_table(self.cols_components, self.data_components)
        
    def print_composite(self, components:list=None):
        self._print_table(self.cols_composite, self.data_composite, components=components)
        
    def print(self):
        print("Components:")
        print("")
        self._print_table(self.cols_components, self.data_components)
        print("")
        print("Composite:")
        if self.norm_type == Norm.NUMBER:
            print("")
            print(f"Total mass:     {self.composite_mass}")
            print(f"Total number:   {self.proportion_norm}")
        print("")
        self._print_table(self.cols_composite, self.data_composite)
        if self.mass_density:
            print("")
            Matter._print(self)


import json

from .node_base import BaseNode
from ..datatypes import StringType

class DescriptionNode(BaseNode):
    keyword: str = 'description'

    @staticmethod
    def is_node(parser):
        parser.kwd_description()
        if parser.is_parsed('kwd_description'):
            parser.part_value()
            parser.part_comment()
            return DescriptionNode(parser)
                     
    def parse(self, env):
        if env.nodes[-1].keyword not in ['str','int','float','bool']:
            raise Exception("Description can be set only to str, int, float and bool nodes:", env.nodes[-1].code)
        if env.nodes[-1].description == None:
            env.nodes[-1].description = str(self.value_raw)
        else:
            env.nodes[-1].description += str(self.value_raw)
        return None



from .type import Type
from .type_boolean import BooleanType
from .type_number import NumberType
from .type_float import FloatType
from .type_integer import IntegerType
from .type_string import StringType



import numpy as np
from dataclasses import dataclass, field, fields
from typing import Union
from math import isclose

from .settings import *

class Fraction:
    __slots__ = ("num","den")

    num: int    # numerator
    den: int    # denominator
    
    @staticmethod
    def from_string(value: str):
        if SYMBOL_FRACTION in value:
            num, den = value.split(SYMBOL_FRACTION)
            return Fraction(int(num), int(den))
        else:
            return Fraction(int(value), 1)

    @staticmethod
    def from_tuple(value: tuple):
        return Fraction(value[0],value[1])

    @staticmethod
    def from_fraction(value: 'Fraction'):
        return Fraction(value.num, value.den)

    def __init__(self, num: int = 0, den: int = 1):
        # ensure whole numbers
        self.num = int(num)
        self.den = int(den)

    def __eq__(self, other):
        return isclose(
            self.num*other.den, 
            other.num*self.den, 
            abs_tol=MAGNITUDE_PRECISION
        )

    def __str__(self):
        self.rebase()
        if self.num==0 or self.den==1:
            return str(self.num)
        else:
            return f"{self.num}{SYMBOL_FRACTION}{self.den}"

    def __repr__(self):
        self.rebase()
        if self.num==0 or self.den==1:
            return str(self.num)
        else:
            return f"{self.num}{SYMBOL_FRACTION}{self.den}"        
        
    def __add__(self, other):
        if isinstance(other, tuple):
            other = Fraction.from_tuple(other) 
        elif isinstance(other, int):
            other = Fraction(other)
        return Fraction(self.num*other.den+other.num*self.den, self.den*other.den)
            
    def __sub__(self, other):
        if isinstance(other, tuple):
            other = Fraction.from_tuple(other) 
        elif isinstance(other, int):
            other = Fraction(other)
        return Fraction(self.num*other.den-other.num*self.den, self.den*other.den)

    def __mul__(self, other):
        if isinstance(other, Fraction):
            return Fraction(self.num*other.num, self.den*other.den)
        elif isinstance(other, tuple):
            return Fraction(self.num*other[0], self.den*other[1])
        else:
            return Fraction(self.num*other, self.den)

    def __truediv__(self, other):
        if isinstance(other, Fraction):
            return Fraction(self.num*other.den, self.den*other.num)
        elif isinstance(other, tuple):
            return Fraction(self.num*other[1], self.den*other[0])
        else:
            return Fraction(self.num, self.den*other)
    
    def __neg__(self):
        return Fraction(-self.num, self.den)

    def rebase(self):
        # reset zero values
        if self.num in [0, -0]:
            self.num = 0
            self.den = 1
        # keep minus sign always on the top
        if self.num>=0 and self.den<0:
            self.num = -self.num
            self.den = -self.den
        elif self.num<0 and self.den<0:
            self.num = -self.num
            self.den = -self.den
        # remove common divisors
        def reduce(num: int, den:int):
            gcd=np.gcd(num, den)
            if gcd>1:
                return reduce(int(num/gcd), int(den/gcd))
            return int(num), int(den)
        self.num, self.den = reduce(self.num, self.den)
    
    def value(self, dtype=tuple):
        if self.num==0 or self.den==1:
            return self.num
        elif dtype==tuple:
            self.rebase()
            return (self.num,self.den)
        elif dtype==float:
            return self.num/self.den



# default DIP parser
from .export         import ExportConfig          
# low-level languages
from .export_c       import ExportConfigC
from .export_cpp     import ExportConfigCPP
from .export_rust    import ExportConfigRust
from .export_fortran import ExportConfigFortran
# parameter formats
from .export_bash    import ExportConfigBash
from .export_json    import ExportConfigJSON
from .export_toml    import ExportConfigTOML
from .export_yaml    import ExportConfigYAML


import numpy as np

from .export import ExportConfig
from ..datatypes import StringType, BooleanType, NumberType, FloatType, IntegerType
from ..settings import Sign

class ExportConfigFortran(ExportConfig):

    def _parse_dtype(self, param, value):
        if isinstance(param, StringType):
            dtype = f"character(len={len(value):d})"
        elif isinstance(param, BooleanType):
            dtype = "logical"
        elif isinstance(param, IntegerType):
            if param.precision==16: dtype = "integer(kind=2)"
            elif param.precision==32: dtype = "integer"
            elif param.precision==64: dtype = "integer(kind=8)"
        elif isinstance(param, FloatType):
            if param.precision==32:    dtype = "real"
            elif param.precision==64:  dtype = "real(kind=8)"
            elif param.precision==128:  dtype = "real(kind=16)"
        return dtype

    def _parse_scalar(self, param, value):
        if isinstance(param, StringType):
            value = f"\"{str(value)}\""
        elif isinstance(param, BooleanType):
            value = ".true." if value else ".false."
        elif isinstance(param, IntegerType):
            value = str(value)
        elif isinstance(param, FloatType):
            value = str(value)
        return value
    
    def _parse_array(self, param, values):
        strings = []
        for v, value in enumerate(values):
            if isinstance(value,(np.ndarray,tuple,list)):
                string, shape = self._parse_array(param, value)
            else:
                string, shape = self._parse_scalar(param, value), None
            strings.append( string )
        if shape is None:
            shape = [len(values)]
        else:
            shape = [len(values)]+shape
        return ", ".join(strings), shape
        
    def _parse_value(self, param, value):
        if isinstance(value,(np.ndarray,tuple,list)):
            return self._parse_array(param, value)
        else:
            return self._parse_scalar(param, value), None

    def parse(self, module:str="ConfigurationModule"):
        lines = []
        lines.append(f"module {module}")
        lines.append("  implicit none")
        lines.append("")
        for name, param in self.data.items():
            name = self._rename(name)
            value, shape = self._parse_value(param, param.value)
            dtype = self._parse_dtype(param, value)
            if shape is None:
                lines.append(f"  {dtype}, parameter :: {name} = {value};")
            else:
                if len(shape)>1:
                    dims = ",".join(str(s) for s in shape)
                    lines.append(f"  {dtype}, dimension ({dims}), parameter :: {name} = reshape([{value}],[{dims}])")
                else:
                    shape = ",".join(str(s) for s in shape)
                    lines.append(f"  {dtype}, dimension ({shape}) :: {name} = [{value}];")
        lines.append("")
        lines.append(f"end module {module}")
        self.text = Sign.NEWLINE.join(lines)
        return self.text


from typing import Union
import numpy as np

class AtomBase:
    
    value: Union[float,bool]

    def __init__(self, value:Union[str,float,bool]):
        if isinstance(value,str):
            self.value = float(value.strip())
        else:
            self.value = value
            
    def __repr__(self):
        return f"Atom({self.value})"
    
    def __add__(self, other):
        return AtomBase(self.value + other.value)
    
    def __sub__(self, other):
        return AtomBase(self.value - other.value)
    
    def __mul__(self, other):
        return AtomBase(self.value * other.value)
    
    def __truediv__(self, other):
        return AtomBase(self.value / other.value)

    def __pow__(self, other):
        return AtomBase(self.value ** other.value)
    
    def __neg__(self):
        return AtomBase(-self.value)

    def log(self):
        return AtomBase(np.log(self.value))

    def log10(self):
        return AtomBase(np.log10(self.value))

    def sqrt(self):
        return AtomBase(np.sqrt(self.value))

    def sin(self):
        return AtomBase(np.sin(self.value))

    def cos(self):
        return AtomBase(np.cos(self.value))

    def tan(self):
        return AtomBase(np.tan(self.value))

    def logical_and(self, other):
        return AtomBase(self.value and other.value)

    def logical_or(self, other):
        return AtomBase(self.value or other.value)

    def logical_not(self):
        return AtomBase(not bool(self.value))

    def __eq__(self, other):
        return AtomBase(self.value == other.value)

    def __ne__(self, other):
        return AtomBase(self.value != other.value)

    def __le__(self, other):
        return AtomBase(self.value <= other.value)

    def __ge__(self, other):
        return AtomBase(self.value >= other.value)

    def __lt__(self, other):
        return AtomBase(self.value < other.value)

    def __gt__(self, other):
        return AtomBase(self.value > other.value)

    



import json

from .settings import *
from .rst_parser import ParseRST
from ..item_parameter import ParType
from ..documentation import Documentation

class ParametersSection:
        
    docs: Documentation
    rst: ParseRST
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass
    
    def __init__(self, docs: Documentation, **kwargs):
        self.docs = docs
        self.rst = ParseRST()

    def styles(self):
        return None
        
    def build_types(self):
        
        self.rst.title(1,'Node types')
        text = ""
        for ptype in ParType:
            text += f"| {ptype.name} - {NTYPES[ptype.value][0]}\n"
        self.rst.paragraph(text)
            
    def build_parameters(self):
        
        self.rst.title(1,'Parameter list')
        header = ['Property name']
        for ptype in ParType:
            th = str(ptype.name)
            header.append(f"{th}")
        data = [header]
        pnames = list(self.docs.parameters.keys())
        pnames.sort()
        for pname in pnames:
            pdata = self.docs.parameters[pname]
            row = [self.rst.link(pdata.target,pname)] + [c if c else '' for c in pdata.counts]
            data.append(row)
            
        self.rst.table_csv(data, **{'widths':'50 5 5 5 5 5 5 5', 'header-rows': 1})

    def build_nodes(self):
        
        self.rst.title(1,'Parameter nodes')
        
        pnames = list(self.docs.parameters.keys())
        pnames.sort()
        for pname in pnames:
            pdata = self.docs.parameters[pname]
            self.rst.target(pdata.target)
            self.rst.title(2, pname)
            for ndata in pdata.nodes:
                source = self.rst.link(ndata.link_source, f"{ndata.source[0]}:{ndata.source[1]}", source=True)
                if ndata.injection:
                    source += " | "+self.rst.link(ndata.link_injection,"injection")
                if ndata.imported:
                    source += " | "+self.rst.link(ndata.link_import,"import")
                ntype = str(list(ParType)[ndata.ntype].name)
                dtype = 'constant '+ndata.dtype if ndata.constant else ndata.dtype
                merge = [] 
                data = [[ntype,source,dtype]]
                if ndata.value:
                    merge.append((len(data),1,len(data),2))
                    data.append(['**Value:**', f"``{ndata.value}``", ''])
                if ndata.unit:
                    merge.append((len(data),1,len(data),2))
                    data.append(['**Unit:**', f"``{ndata.unit}``", ''])
                if ndata.condition:
                    merge.append((len(data),1,len(data),2))
                    data.append(['**Condition:**', f"``{ndata.condition}``", ''])
                if ndata.options:
                    merge.append((len(data),1,len(data),2))
                    data.append(['**Options:**', ", ".join(ndata.options), ''])
                if ndata.tags:
                    merge.append((len(data),1,len(data),2))
                    data.append(['**Tags:**', ", ".join(ndata.tags), ''])
                if ndata.dformat:
                    merge.append((len(data),1,len(data),2))
                    data.append(['**Format:**', f"``{ndata.dformat}``", ''])
                if ndata.description:
                    merge.append((len(data),1,len(data),2))
                    data.append(['**Description:**', ndata.description, ''])
                self.rst.target(ndata.target)
                self.rst.table(data, merge=merge, header=True)

    def build(self):
        
        self.build_types()
        
        self.build_parameters()
        
        self.build_nodes()

        return self.rst.parse()


from reportlab.platypus import SimpleDocTemplate, Paragraph, Spacer, Table, KeepTogether, PageBreak
from reportlab.platypus.doctemplate import PageTemplate, BaseDocTemplate
from reportlab.platypus.tableofcontents import TableOfContents
from reportlab.platypus.flowables import KeepTogether 
from reportlab.platypus.frames import Frame
from reportlab.lib.styles import ParagraphStyle
from reportlab.lib.units import cm
import numpy as np
from dataclasses import dataclass

from .settings import *
from .node import NodeSection
from .injections import InjectionsSection
from .imports import ImportsSection
from .types import TypesSection
from .parameters import ParametersSection
from .units import UnitsSection
from .sources import SourcesSection
from ..documentation import Documentation
from ..settings import DocsType
from ...lists import NodeList
from ...nodes import Node, BooleanNode, IntegerNode, FloatNode, StringNode, ModNode, ImportNode

def pageSettings(canvas, doc):
    canvas.saveState()
    canvas.setFont('Times-Roman',9)
    canvas.drawString(2.5*cm, 2*cm, "Page %d " % (doc.page))
    canvas.restoreState()

class DocsTemplate(BaseDocTemplate):    
    def __init__(self, filename, **kw):
        self.allowSplitting = 0
        BaseDocTemplate.__init__(self, filename, **kw)
        template = PageTemplate('normal', [Frame(2.5*cm, 2.5*cm, 15*cm, 25*cm, id='F1')], onPage=pageSettings)
        self.addPageTemplates(template)

    def afterFlowable(self, flowable):
        "Registers TOC entries."
        if flowable.__class__.__name__ == 'Paragraph':
            text = Link(f"SECTION_{flowable.getPlainText()}", flowable.getPlainText())
            style = flowable.style.name
            if style == 'Heading1':
                self.notify('TOCEntry', (0, text, self.page))
            if style == 'Heading2':
                self.notify('TOCEntry', (1, text, self.page))
                
class ExportDocsPDF:
    
    docs: Documentation
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass
    
    def __init__(self, docs: Documentation, **kwargs):
        self.docs = docs
        
    def build(self, file_path: str, title, intro=None):

        blocks = []

        blocks.append(Paragraph(title, TITLE))
        
        blocks.append(Paragraph(Title("Table of contents"), H0))
        
        toc = TableOfContents()
        toc.levelStyles = [
            ParagraphStyle(name='TOCHeading1', fontSize=16, leftIndent=20, firstLineIndent=-20, spaceBefore=5, leading=16, fontName='Times-Bold'),
            ParagraphStyle(name='TOCHeading2', fontSize=14, leftIndent=40, firstLineIndent=-20, spaceBefore=5, leading=12),
        ]
        blocks.append(toc)
    
        if intro:
            blocks.append(PageBreak())
            blocks.append(Paragraph(Title("Introduction"), H1))
            blocks.append(Paragraph(intro))
            blocks.append(Spacer(1,cm))
        
        blocks.append(PageBreak())
        blocks.append(Paragraph(Title("Parameters"), H1))
        
        # list node types
        with TypesSection(self.docs.types) as tmpl:
            blocks += tmpl.parse()    

        # create a quick reference of nodes with links
        with ParametersSection(self.docs.parameters) as tmpl:
            blocks += tmpl.parse()
        
        blocks.append(PageBreak())

        # list all nodes and their properties
        with NodeSection(self.docs.parameters) as tmpl:
            blocks += tmpl.parse()
            
        blocks.append(PageBreak())
        blocks.append(Paragraph(Title("References"), H1))

        # list all injected nodes
        with InjectionsSection(self.docs.injections) as tmpl:
            blocks += tmpl.parse()
            
        blocks.append(PageBreak())

        # list all imports
        with ImportsSection(self.docs.imports) as tmpl:
            blocks += tmpl.parse()
            
        blocks.append(PageBreak())
        blocks.append(Paragraph(Title("Settings"), H1))
        
        # list units
        with UnitsSection(self.docs.units) as tmpl:
            blocks += tmpl.parse()    
 
        # list sources
        blocks.append(PageBreak())
        with SourcesSection(self.docs.sources) as tmpl:
            blocks += tmpl.parse()    
 
        doc = DocsTemplate(file_path)
        doc.multiBuild(blocks)
            


from bs4 import BeautifulSoup
import json

from .settings import *
from ..item_parameter import ParType
from ..documentation import Documentation

class ReferencesSection:
        
    docs: Documentation
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass
    
    def __init__(self, docs: Documentation, **kwargs):
        self.docs = docs
        self.html = BeautifulSoup(f"", 'html.parser')
        
    def styles(self):
        return None
        
    def build_injections(self):
        section = BeautifulSoup(Title("Injections",3), 'html.parser')
        self.html.append(section)
        
        def add_prop(name, value, append=False):
            pname = self.html.new_tag('div', **{'class':'col-md-3 bg-body-secondary'})
            props = self.html.new_tag('div', **{'class':'row bg-white'})
            pname.string = name
            props.append(pname)
            pvalue = self.html.new_tag('div', **{'class':'col'})
            if append:
                pvalue.append(value)
            else:
                pvalue.string = value
            props.append(pvalue)
            return props
        
        for idata in self.docs.injections:
            item = self.html.new_tag("div", **{'class':"container mt-3 border"})
            
            header = self.html.new_tag('div', **{'class':'row bg-dark-subtle'})
            name = self.html.new_tag('div', **{'class':'col'})
            name.append(Target(idata.target))
            name.append(Link(PAGE_SOURCES, idata.link_source, f"{idata.source[0]}:{idata.source[1]}"))
            header.append(name)
            item.append(header)
            
            item.append(add_prop('Injectiong node:', Link(PAGE_NODES,idata.link_node, idata.name), append=True))
            item.append(add_prop('Request:', "{"+idata.reference+"}"))
            if idata.isource:
                item.append(add_prop('From source:', Link(PAGE_SOURCES,idata.link_isource, f"{idata.isource[0]}:{idata.isource[1]}"), append=True))
            if idata.ivalue:
                item.append(add_prop('Value:', idata.ivalue))
            self.html.append(item)
        
    def build_imports(self):
        section = BeautifulSoup(Title("Imports",3), 'html.parser')
        self.html.append(section)
        
        for idata in self.docs.imports:
            item = self.html.new_tag("div", **{'class':"container mt-3 border"})
            
            header = self.html.new_tag('div', **{'class':'row bg-dark-subtle'})
            name = self.html.new_tag('div', **{'class':'col'})
            name.append(Target(idata.target))
            name.append(Link(PAGE_SOURCES,idata.link_source,f"{idata.source[0]}:{idata.source[1]}"))
            header.append(name)
            item.append(header)
            
            props = self.html.new_tag('div', **{'class':'row'})
            pname = self.html.new_tag('div', **{'class':'col-md-2 bg-body-secondary'})
            pname.string = 'Request:'
            props.append(pname)
            pvalue = self.html.new_tag('div', **{'class':'col bg-white'})
            pvalue.string = "{"+idata.reference+"}"
            props.append(pvalue)
            item.append(props)
            
            if idata.idata:
                props = self.html.new_tag('div', **{'class':'row'})
                pname = self.html.new_tag('div', **{'class':'col-6 bg-body-secondary'})
                pname.string = 'Imported node:'
                props.append(pname)
                pvalue = self.html.new_tag('div', **{'class':'col-6 bg-body-secondary'})
                pvalue.string = "From source:"
                props.append(pvalue)
                item.append(props)
            
            for inode in idata.idata:
                props = self.html.new_tag('div', **{'class':'row'})
                pname = self.html.new_tag('div', **{'class':'col-6 bg-white'})
                pname.append(Link(PAGE_NODES, inode.link_node, inode.name))
                props.append(pname)
                pvalue = self.html.new_tag('div', **{'class':'col-6 bg-white'})
                pvalue.append(Link(PAGE_SOURCES, inode.link_source, f"{inode.source[0]}:{inode.source[1]}"))
                props.append(pvalue)
                item.append(props)
            
            self.html.append(item)        
            
    def build(self):
        
        self.build_injections()
        
        self.build_imports()

        return self.html


import sys, os
sys.path.insert(0, os.environ['DIR_SOURCE'])

from scinumtools.units.settings import *
from scinumtools.units.unit_solver import UnitSolver
from scinumtools.units.base_units import BaseUnits

def build_unit_systems():

    text = [
        "#############################################",
        "# Do not modify this file!                  #",
        "# It is generated automatically in:         #",
        "# tools/generator/build_units.py            #",
        "#############################################",
        ""
        "QUANTITY_UNITS = {",
    ]
    symbols = []
    for q in range(len(QUANTITY_LIST)):
        name = QUANTITY_LIST.name[q]
        text.append(f"  # {name}")
        for system in ['SI','CGS','AU']:
            definition = getattr(QUANTITY_LIST,system)[q]
            if definition is None:
                continue
            else:
                symbol = QUANTITY_LIST.symbol[q]
                unitid = f"#{system[0]}{symbol}"
                atom = UnitSolver(definition)
                base = BaseUnits(atom.baseunits)
                c1 = f"{atom.magnitude*base.magnitude},"
                c2 = f"{base.dimensions.value()}"
                symbols.append(unitid)
                text.append(f"  '{unitid}': ({c1:25s}{c2:27s}), # {definition}") 
    text.append("}")
    text = "\n".join(text)
    
    # test if new symbols are unique
    try:
        assert len(np.unique(symbols)) == len(symbols)
    except:
        unique = np.unique(symbols)
        notunique = [s for s in unique if symbols.count(s)>1]
        raise Exception("Some units symbols are not unique:", notunique)
    
    # test if we produced a valid Python code
    exec(text)
    assert QUANTITY_UNITS
    
    # save the new version of the code
    path_units = os.environ['DIR_SOURCE']+"/scinumtools/units"
    path_list = f'{path_units}/unit_list.py'
    assert os.path.isdir(path_units)
    with open(path_list,'w') as f:
        f.write(text)
    print(path_list)


import numpy as np

from .export import ExportConfig
from ..datatypes import StringType, BooleanType, NumberType, FloatType, IntegerType
from ..settings import Sign

class ExportConfigRust(ExportConfig):
    
    def _parse_dtype(self, param, value):
        if isinstance(param, StringType):
            dtype = "&str"
        elif isinstance(param, BooleanType):
            dtype = "bool"
        elif isinstance(param, IntegerType):
            if param.unsigned:
                if param.precision==8:    dtype = "u8"
                elif param.precision==16: dtype = "u16"
                elif param.precision==32: dtype = "u32"
                elif param.precision==64: dtype = "u64"
            else:
                if param.precision==8:    dtype = "i8"
                elif param.precision==16: dtype = "i16"
                elif param.precision==32: dtype = "i32"
                elif param.precision==64: dtype = "i64"
        elif isinstance(param, FloatType):
            if param.precision==32:    dtype = "f32"
            elif param.precision==64:  dtype = "f64"
            elif param.precision==128: dtype = "f64"  # f128 is not supported yet
        return dtype
    
    def _parse_scalar(self, param, value):
        if isinstance(param, StringType):
            value = f"\"{value}\""
        elif isinstance(param, BooleanType):
            value = "true" if value else "false"
        elif isinstance(param, IntegerType):
            value = str(value)
        elif isinstance(param, FloatType):
            value = str(value)
        return value
    
    def _parse_array(self, param, values):
        strings = []
        for v, value in enumerate(values):
            if isinstance(value,(np.ndarray,tuple,list)):
                string, shape = self._parse_array(param, value)
            else:
                string, shape = self._parse_scalar(param, value), None
            strings.append( string )
        if shape is None:
            shape = [len(values)]
        else:
            shape = [len(values)]+shape
        return "[" + ", ".join(strings) + "]", shape
        
    def _parse_value(self, param, value):
        if isinstance(value,(np.ndarray,tuple,list)):
            return self._parse_array(param, value)
        else:
            return self._parse_scalar(param, value), None
            
    def parse(self):
        lines = []
        for name, param in self.data.items():
            name = self._rename(name)
            value, size = self._parse_value(param, param.value)
            dtype = self._parse_dtype(param, param.value)
            if size is not None:
                size.reverse()
                for dim in size:
                    dtype = f"[{dtype}; {dim}]"
            lines.append(f"pub const {name}: {dtype} = {value};")
        self.text = Sign.NEWLINE.join(lines)
        return self.text



import numpy as np
import re
from math import isclose
import copy
from typing import Union

from . import Norm, Units
from .matter import Matter
from .composite import Composite, Component
from .element import Element
from .substance_solver import SubstanceSolver
from .. import ParameterTable, RowCollector
from ..units import Quantity, Unit

class Substance(Composite, Component, Matter):
    
    def atom(self, expr:str):
        if m:=re.match("[0-9]+(\.[0-9]+|)([eE]{1}[+-]?[0-9]{0,3}|)",expr):
            return float(expr)
        else:
            return Substance({expr: 1}, natural=self.natural)
            
    def __init__(self, expr:Union[str,dict]=None, proportion:float=1.0, natural:bool=True, **kwargs):
        Matter.__init__(self, **kwargs)
        Component.__init__(self, proportion)
        Composite.__init__(self, 
            SubstanceSolver, Element, expr, {
                'element':    None,
                'isotope':    None,
                'ionisation': None,
                'mass':       Units.ATOMIC_MASS,
                'count':      None,
                'Z':          None,
                'N':          None,
                'e':          None,
            }, {
                "mass":  Units.ATOMIC_MASS, 
                'Z':     None,
                'N':     None,
                'e':     None,
            }, natural=natural, norm_type=Norm.NUMBER
        )

    def __str__(self):
        elements = []
        for expr, el in self.components.items():
            element = f"{expr}" 
            if el.proportion>1:
                element += f"{int(el.proportion):d}"
            elements.append(element)
        elements = " ".join(elements)
        data = self.data_composite(quantity=False)
        p = int(round(data['sum']['Z']))
        n = data['sum']['N']
        e = int(round(data['sum']['e']))
        M = data['sum']['mass']
        return f"Substance(mass={M:.03f} Z={p:d} N={n:.03f} e={e:d})"
        
    def __mul__(self, other:float):
        return self._multiply(Substance(natural=self.natural), other)

    def __add__(self, other):
        return self._add(Substance(natural=self.natural), other)
    
    def _add_expr(self, expr:str, proportion:int):
        self.expr += f"{expr}{proportion}" if proportion>1 else expr
    
    def data_components(self, quantity:bool=True):
        def fn_row(s,m):
            return {
                'element':    m.element, 
                'isotope':    m.isotope, 
                'ionisation': m.ionisation, 
                'mass':       m.mass, 
                'count':      m.proportion, 
                'Z':          m.Z, 
                'N':          m.N, 
                'e':          m.e, 
            }
        return self._data(self.cols_components, fn_row, quantity=quantity)
        
    def data_composite(self, components:list=None, quantity:bool=True):
        def fn_row(s,m):
            return {
                'mass':  m.proportion*m.mass,
                'Z':     m.proportion*m.Z, 
                'N':     m.proportion*m.N, 
                'e':     m.proportion*m.e, 
            }
        return self._data(self.cols_composite, fn_row, stats=True, weight=True, components=components, quantity=quantity)


import re

from .settings import *
from ..solver import ExpressionSolver, OperatorPar, OperatorMul, OperatorTruediv
from .fraction import Fraction

class Atom:
    
    magnitude: float
    baseunits: dict
    
    def __init__(self, magnitude, baseunits):
        self.magnitude = magnitude
        self.baseunits = baseunits

    def __mul__(self, other):
        magnitude = self.magnitude * other.magnitude
        baseunits = dict(self.baseunits)
        for unit,exp in other.baseunits.items():
            baseunits[unit] = baseunits[unit]+exp if unit in baseunits else exp
        return Atom(magnitude, baseunits)
    
    def __truediv__(self, other):
        magnitude = self.magnitude / other.magnitude
        baseunits = dict(self.baseunits)
        for unit,exp in other.baseunits.items():
            baseunits[unit] = baseunits[unit]-exp if unit in baseunits else -exp
        return Atom(magnitude, baseunits)
        
    def __str__(self):
        baseunits = []
        for unit,exp in self.baseunits.items():
            if exp.num not in [0, -0]:
                unit = unit.replace(SYMBOL_UNITID,"")
                baseunits.append(f"{unit}={str(exp)}")
        baseunits = " ".join(baseunits)
        if baseunits:
            return f"Atom({self.magnitude:.3e} {baseunits})"
        else:
            return f"Atom({self.magnitude:.3e})"
            
    def __repr__(self):
        baseunits = []
        for unit,exp in self.baseunits.items():
            if exp.num not in [0, -0]:
                unit = unit.replace(SYMBOL_UNITID,"")
                baseunits.append(f"{unit}={str(exp)}")
        baseunits = " ".join(baseunits)
        if baseunits:
            return f"Atom({self.magnitude:.3e} {baseunits})"
        else:
            return f"Atom({self.magnitude:.3e})"

def AtomParser(string=None):
    # parse number
    if m := re.match(r'^[-]?([0-9.]+)(e([0-9+-]+)|)$', str(string)):
        magnitude = float(string)
        return Atom(magnitude, {})
    # parse unit
    string_bak = string
    string = ' '+string
    # parse exponent
    if m := re.search(r"[0-9"+SYMBOL_FRACTION+r"+-]+$", string):
        exp = m.group() 
        string = string[:-len(exp)]
        exp = Fraction.from_string(exp)
    else:
        exp = Fraction(1)
    if string.startswith(f" {SYMBOL_SYSTEM_UNIT}"):
        unitid = string[1:]
        return Atom(1.0, {unitid: exp})
    # parse unit symbol
    bases = [u for u in UNIT_STANDARD.keys() if string.endswith(u)]
    if bases:
        base = max(bases, key=len)
        string = string[-len(base)-1]
        unitid = f"{base:s}"
    else:
        raise Exception('Unknown unit', string, string_bak)
    # parse unit prefix
    prefkeys = [p for p in UNIT_PREFIXES.keys() if string.endswith(p)]
    if prefkeys:
        prefix = max(prefkeys, key=len)
        if isinstance(UNIT_STANDARD[base].prefixes,list) and prefix not in UNIT_STANDARD[base].prefixes:
            raise Exception(f"Unit can have only following prefixes:", UNIT_STANDARD[base].prefixes, prefix)
        elif UNIT_STANDARD[base].prefixes is True and prefix not in UNIT_PREFIXES.keys():
            raise Exception(f"Unknown unit prefix:", string_bak)
        elif UNIT_STANDARD[base].prefixes is False:
            raise Exception(f"Unit cannot have any prefixes:", base)
        unitid = f"{prefix:s}{SYMBOL_UNITID}{unitid}"
    elif len(string)>1:
        raise Exception("Unknown unit prefix:", string)
    # return quantity
    return Atom(1.0, {unitid: exp})
        
def UnitSolver(expression):
    operators = {'par':OperatorPar,'mul':OperatorMul,'truediv':OperatorTruediv}
    with ExpressionSolver(AtomParser, operators) as es:
        return es.solve(expression)


import numpy as np
import re
import os
from pathlib import Path
from typing import List, Callable, Tuple
from inspect import getframeinfo, stack

from .environment import Environment
from .docs import Documentation
from .settings import *
from .docs.settings import DocsType
from .nodes.parser import Parser
from .nodes import EmptyNode, ImportNode, UnitNode, SourceNode, CaseNode
from .nodes import OptionNode, ConstantNode, FormatNode, ConditionNode, TagsNode, DescriptionNode
from .nodes import ModNode, GroupNode
from .nodes import BooleanNode, IntegerNode, FloatNode, StringNode, TableNode
from .solvers import LogicalSolver
from .datatypes import Type

class DIP:
    """ DIP parser class

    :param str code: DIP code
    :param DIP_Environment env: DIP environment object
    """
    name: str            # object name
    env: Environment     # environment
    lines: List[dict]    # code lines
    
    source: tuple        # source

    nodes_special: list     = ['empty','unit','source']
    nodes_properties: list  = ['option','constant','format','condition','tags','description']
    nodes_hierarchy: list   = ['group']
    nodes_nohierarchy: list = None
    nodes_notypes: list     = None
    
    num_files: int = 0 
    num_strings: int = 0

    def __init__(self, env:Environment=None, **kwargs):
        self.name = kwargs['name'] if 'name' in kwargs else str(id(self))
        self.lines = []
        # create a new environment if not givenl
        if env:
            self.env = env
        elif 'env' not in kwargs:
            self.env = Environment()
        else:
            self.env = None
        # find out source
        if 'source' not in kwargs:
            # determine which file instantiate this class
            caller = getframeinfo(stack()[1][0])
            self.source = (f"{self.name}_{ROOT_SOURCE}", None)
            if caller.filename == '<stdin>':                
                self.env.sources.append(
                    name = self.source[0], 
                    path = os.getcwd(),
                    code = None,
                )
            else:
                self.env.sources.append(
                    name = self.source[0], 
                    path = caller.filename, 
                    code = None, 
                )
        else:
            self.source = kwargs['source']
        # create additional node groups
        self.nodes_nohierarchy = self.nodes_special+self.nodes_properties
        self.nodes_notypes = self.nodes_special+self.nodes_properties+self.nodes_hierarchy

    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass
        
    def _get_queue(self):
        # Convert code lines to nodes
        queue = Environment()
        while len(self.lines)>0:
            line = self.lines.pop(0)
            # Group block structures
            if '"""' in line['code']:
                block = []
                while len(self.lines)>0:
                    subline = self.lines.pop(0)
                    if '"""' in subline['code']:
                        line['code'] += Sign.NEWLINE.join(block) + subline['code'].lstrip()
                        break
                    else:
                        block.append( subline['code'] )
                else:
                    raise Exception("Block structure is not properly terminated.", line['code'])
            node = self._determine_node(line)
            queue.nodes.append(node)
        return queue

    def _determine_node(self, line):
        # Add replacement marks
        # TODO: we need to also properly treate arrays like this ["d#", "b"]
        encode = ["\\'", '\\"', "\n"]
        for i,symbol in enumerate(encode):
            line['code'] = line['code'].replace(symbol,f"$@{i:02d}")
            
        # Determine node type
        parser = Parser(
            code=line['code'],
            source=line['source'],
        )
        steps = [
            EmptyNode.is_node,            # parse empty line node
            parser.part_indent,           
            ImportNode.is_node,           # parse root import directive
            UnitNode.is_node,             # parse unit directive
            SourceNode.is_node,           # parse source directive
            CaseNode.is_node,             # parse case directive
            OptionNode.is_node,           # parse option setting
            ConstantNode.is_node,         # parse constant setting
            FormatNode.is_node,           # parse format setting
            TagsNode.is_node,             # parse node tags
            DescriptionNode.is_node,      # parse description node
            ConditionNode.is_node,        # parse condition setting
            parser.part_name,             
            GroupNode.is_node,            # parse group node
            ImportNode.is_node,           # parse group import directive
            ModNode.is_node,              # parse modification
            parser.part_type,             
            BooleanNode.is_node,          # parse boolean node
            IntegerNode.is_node,          # parse integer node
            FloatNode.is_node,            # parse float node
            StringNode.is_node,           # parse string node
            TableNode.is_node,            # parse table node
        ]
        node = None
        for step in steps:
            if step.__name__=='is_node':
                node = step(parser)
                if node:
                    node = node
            else:
                step()
            if node and parser.is_empty():
                break
        else:
            raise Exception(f"Code cannot be parsed:",parser.ccode)
        
        # Convert symbols to original letters
        def decode_symbols(value):
            replace = ["\'", '\"', "\n"]
            if isinstance(value, (list, np.ndarray)):
                value = [decode_symbols(v) for v in value]
            elif value is None:
                return value
            else:
                for i,symbol in enumerate(replace):
                    value = value.replace(f"$@{i:02d}", symbol)
            return value
        
        # Remove replacement marks
        if node.value_fn:
            node.value_fn = decode_symbols(node.value_fn)
        if node.value_expr:
            node.value_expr = decode_symbols(node.value_expr)
        node.value_raw = decode_symbols(node.value_raw)
        node.code = decode_symbols(node.code)

        # Return proper node type
        return node
    
    def add_file(self, filepath:str, source_name:str = None, absolute=True):
        """ Load DIP code from a file

        :param str filepath: Path to a DIP file
        
        :param bool absolute: Path is absolute
        """
        # ensure relative paths with respect to the source script
        if not os.path.isabs(filepath) and absolute:
            parent = Path(self.env.sources[self.source[0]].path)
            if os.path.isfile(parent):
                filepath = parent.parent / filepath
            else:
                filepath = parent / filepath
        # open file and get its content
        with open(filepath,'r') as f:           
            lines = f.read().split(Sign.NEWLINE)
        # strip leading and ending empty lines
        while lines and lines[0].strip()=='':
            del lines[0]
        while lines and lines[-1].strip()=='':
            del lines[-1]
        code = Sign.NEWLINE.join(lines)
        # get source name
        if source_name is None:
            self.num_files += 1
            source_name = f"{self.name}_{FILE_SOURCE}{self.num_files}"
        # prepare individual lines
        for lineno, linecode in enumerate(lines):
            self.lines.append(dict(
                code = linecode,
                source = (source_name, lineno+1),
            ))
        # get line number from the source code
        if self.source[1] is None:
            caller = getframeinfo(stack()[1][0])
            lineno = caller.lineno
        else:
            lineno = self.source[1]
        # create new source
        self.env.sources.append(
            name = source_name,
            path = os.path.realpath(filepath),
            code = Sign.NEWLINE.join(lines),
            parent = (self.source[0], lineno),
        )

    def add_string(self, code:str):
        """ Use DIP code from a string

        :param str code: DIP code
        """
        lines = code.split(Sign.NEWLINE)
        # create source name
        self.num_strings += 1
        source_name = f"{self.name}_{STRING_SOURCE}{self.num_strings}"
        # strip leading and ending empty lines
        while lines and lines[0].strip()=='':
            del lines[0]
        while lines and lines[-1].strip()=='':
            del lines[-1]
        code = Sign.NEWLINE.join(lines)
        # create lines
        for lineno,linecode in enumerate(lines):
            self.lines.append(dict(
                code = linecode,
                source = (source_name, lineno+1),
            ))
        # create parent new source
        if self.source[1] is None:
            caller = getframeinfo(stack()[1][0])
            lineno = caller.lineno
        else:
            lineno = self.source[1]
        self.env.sources.append(
            name = source_name,
            path = self.env.sources[self.source[0]].path,
            code = code,
            parent = (self.source[0], lineno),
        )

    def add_source(self, name:str, path:str):
        if self.source[1] is None:
            caller = getframeinfo(stack()[1][0])
            lineno = caller.lineno
        else:
            lineno = self.source[1]
        self.lines.append(dict(
            code = f"{Sign.VARIABLE}{Keyword.SOURCE} {name} = '{path}'",
            name = name,
            source = (self.source[0], lineno),
        ))
        
    def add_unit(self, name:str, value:float, unit:str=None):
        if unit:
            code = f"{Sign.VARIABLE}{Keyword.UNIT} {name} = {value} {unit}"
        else:
            code = f"{Sign.VARIABLE}{Keyword.UNIT} {name} = {value}"
        if self.source[1] is None:
            caller = getframeinfo(stack()[1][0])
            lineno = caller.lineno
        else:
            lineno = self.source[1]
        self.lines.append(dict(
            code = code,
            name = name,
            source = (self.source[0], lineno),
        ))

    def add_function(self, name:str, fn:Callable):
        self.env.functions.append(name, fn)

    def parse(self):
        """ Parse DIP nodes from code lines
        """
        # Create queue/target environment
        queue = self._get_queue()
        target = self.env.copy()
        # Parse nodes
        while len(queue.nodes):
            node = queue.nodes.pop()
            # Perform specific node parsing only outside of case or inside of valid case
            if not target.branching.false_case() or node.keyword=='case':
                node.inject_value(target)
                parsed = node.parse(target)
                if parsed: 
                    # Add parsed nodes to the queue and continue
                    queue.nodes.prepend(parsed)
                    continue
            # Create hierarchical name
            target.hierarchy.register(node, self.nodes_nohierarchy)
            # Add nodes to the node list
            if node.keyword in self.nodes_notypes:
                continue
            elif node.keyword=='case':   # Parse cases
                target.branching.solve_case(node)
            elif target.branching.false_case():
                continue
            else:
                target.branching.prepare_node(node)
                # Clean node name from cases
                node.name = node.clean_name()
                # Set the node value
                node.set_value()
                # If node was previously defined, modify its value
                for n in range(len(target.nodes)):
                    if target.nodes[n].name==node.name:
                        if target.nodes[n].constant:
                            raise Exception(f"Node '{target.nodes[n].name}' is constant and cannot be modified:",node.code)
                        target.nodes[n].modify_value(node, target)
                        break
                # If node wasn't defined, create a new node
                else:
                    if node.keyword=='mod' and node.source[0].startswith(f"{self.name}_{STRING_SOURCE}"):
                        raise Exception(f"Modifying undefined node:",node.name)
                    target.nodes.append(node)
        # Validate nodes
        for node in target.nodes:
            # Check if all declared nodes have assigned value
            if node.defined and node.value is None:
                raise Exception(f"Node value must be defined:", node.code)
            # Check if node value is in options
            if isinstance(node,(IntegerNode, FloatNode, StringNode)):
                node.validate_options()
            # Check conditions
            if node.keyword in ['float','int'] and node.condition:
                target.autoref = node.name
                with LogicalSolver(target) as s:
                    if not s.solve(node.condition).value:
                        raise Exception("Node does not fullfil a condition:",
                                        node.name, node.condition)
                target.autoref = None
            # Check formats if set for strings
            if node.keyword=='str' and node.format:
                m = re.match(node.format, node.value.value)
                if not m:
                    raise Exception("Node value does not match the format:",
                                    node.value.value, node.format)
        return target
        
    def parse_docs(self):
        """ Parse DIP node definitions for a documentation
        """
        # Create queue/target environment
        queue = self._get_queue()
        target = self.env.copy()
        target.envtype = EnvType.DOCS
        # Parse nodes
        while len(queue.nodes):
            node = queue.nodes.pop()
            # Perform specific node parsing
            node.inject_value(target)
            parsed = node.parse(target)
            if parsed: 
                # Add parsed nodes to the queue and continue
                queue.nodes.prepend(parsed)
                if node.keyword!=ImportNode.keyword:
                    continue
            # Create hierarchical name
            target.hierarchy.register(node, self.nodes_nohierarchy)
            # Add nodes to the node list
            if node.keyword in self.nodes_notypes:
                continue
            elif node.keyword=='case':   # Parse cases
                target.branching.solve_case(node)
            else:
                target.branching.prepare_node(node)
                # Set the node value
                node.set_value()
                # Set empty DocsType
                node.docs_type = DocsType.DEFINITION if node.value else DocsType.DECLARATION
                # Loop through all nodes and find modifications
                for n in range(len(target.nodes)):
                    # Continue if names do not match
                    target_name = target.nodes[n].clean_name()
                    target_branch_id = target.nodes[n].branch_id
                    if target_name != node.clean_name():
                        continue  # nodes have different names
                    if target_branch_id==node.branch_id:
                        if target.nodes[n].case_id!=node.case_id:
                            continue  # nodes are in the same branch but no case
                    if target_branch_id is not None:
                        target_branch = target.branching.branches[target_branch_id]
                        if 'else' in target_branch.types:
                            if target_branch.nodes[target_name] != len(target_branch.cases):
                                node.docs_type |= DocsType.MODIFICATION
                                continue # node was not completely defined in its branch
                        else:
                            node.docs_type |= DocsType.MODIFICATION
                            continue # node branch does not have an else clause 
                    node.docs_type = DocsType.MODIFICATION
                    break
                target.nodes.append(node)
        return Documentation(target)

    def setup(self):
        pass


import numpy as np

from ..settings import Sign, EnvType, Format
from ..datatypes import StringType, BooleanType, NumberType, FloatType, IntegerType
from ..nodes import StringNode, BooleanNode, FloatNode, IntegerNode
from ..environment import Environment

class ExportConfig:
    
    env: Environment
    data: dict
    dtype: Format
    rename: bool
    text: str

    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass
    
    def __init__(self, env: Environment, **kwargs):
        """ Initialize export
        :param Environment env: DIP environment that should be exported
        :param Format dtype: DIP data format 
        """
        if env.envtype != EnvType.DATA:
            raise Exception("Given environment is not a data environment")
        self.env = env
        self.dtype = kwargs['dtype'] if 'dtype' in kwargs else Format.TYPE
        self.rename = kwargs['rename'] if 'rename' in kwargs else True
        self.data = self.env.data(self.dtype)
    
    def _rename(self, name):
        """ Internal routine that renames parameter names into a proper format
        """
        if self.rename:
            return name.upper().replace(Sign.SEPARATOR,"_")
        else:
            return name
    
    def select(self, query:str=None, tags:list=None):
        """ Select nodes from an environment
        :param str query: Node query string
        :param list tags: List of tag selectors
        """
        self.data = self.env.data(self.dtype, query=query, tags=tags)

    def parse(self):
        """ Default DIP parser
        """
        lines = []
        for name, param in self.data.items():
            value = param.value
            if isinstance(param, StringType):
                dtype = StringNode.keyword
                value = f"\"{value}\""
            elif isinstance(param, BooleanType):
                dtype = BooleanNode.keyword
                value = "true" if value else "false"
            elif isinstance(param, IntegerType):
                dtype = IntegerNode.keyword
                if param.unsigned:
                    dtype = "u"+dtype
                if param.precision!=IntegerType.precision:
                    dtype += str(param.precision)
                value = int(param.value)
            elif isinstance(param, FloatType):
                dtype = FloatNode.keyword
                if param.precision!=FloatType.precision:
                    dtype += str(param.precision)
                value = float(param.value)
            if param.unit:
                lines.append(f"{name} {dtype} = {value} {param.unit}")
            else:
                lines.append(f"{name} {dtype} = {value}")
        self.text = Sign.NEWLINE.join(lines)
        return self.text

    def save(self, file_path: str, mode:str = 'w'):
        """ Save parsed text into a file
        """
        with open(file_path, mode) as f:
            f.write(self.text)


import numpy as np
import os
import sys
sys.path.insert(0, os.environ['DIR_SOURCE'])

from scinumtools import RowCollector, ParameterTable
from scinumtools.units import Dimensions
from scinumtools.units.settings import *
from scinumtools.units.unit_types import *
from scinumtools.dip import DIP
from scinumtools.materials.periodic_table import *

path_docs_static = os.environ['DIR_DOCS']+'/source/_static/tables'

def build_prefixes():
    
    # Create table of unit prefixes
    rc = RowCollector(['Symbol','Name','Magnitude'])
    for symbol, prefix in UNIT_PREFIXES.items():
        exponent = int(np.log10(prefix.magnitude))
        rc.append([symbol, prefix.name, ":math:`10^{"+f"{exponent:d}"+"}`"])
    df = rc.to_dataframe()
    file_path = f"{path_docs_static}/prefixes.csv"
    df.to_csv(file_path, index=False)
    print(file_path)

def build_base_units():

    # create table of base units
    rc = RowCollector(['Symbol','Name','Prefixes'])
    for symbol, unit in UNIT_STANDARD.items():
        if unit.definition is not None:
            continue
        if unit.prefixes is True:
            prefixes = 'all'
        elif isinstance(unit.prefixes, list):
            prefixes = ", ".join([f"{p}{symbol}" for p in unit.prefixes])
        else:
            prefixes = ''
        rc.append([symbol, unit.name, prefixes])
    df = rc.to_dataframe()
    file_path = f"{path_docs_static}/unit_base.csv"
    df.to_csv(file_path, index=False)
    print(file_path)
    
def build_standard_units():    

    # create table of standard units
    rc = RowCollector(['Symbol','Name','Definition','Prefixes'])
    for symbol, unit in UNIT_STANDARD.items():
        if not isinstance(unit.definition, str):
            continue
        if symbol in ['K','degR','AR','PR']:
            continue
        if symbol.startswith('['):
            continue
        if unit.prefixes is True:
            prefixes = 'all'
        elif isinstance(unit.prefixes, list):
            prefixes = ", ".join([f"{p}{symbol}" for p in unit.prefixes])
        else:
            prefixes = ''
        if len(rc) and unit.name in rc.Name:
            idx = rc.Name.index(unit.name)
            rc.Symbol[idx] += ", "+symbol
            rc.Prefixes[idx] += ", "+prefixes if rc.Prefixes[idx] else ""
        else:
            rc.append([symbol, unit.name, unit.definition, prefixes])
    df = rc.to_dataframe()
    file_path = f"{path_docs_static}/unit_standard.csv"
    df.to_csv(file_path, index=False)
    print(file_path)
    
def build_constants():
    
    # create table of constants
    rc = RowCollector(['Symbol','Name','Definition'])
    for symbol, unit in UNIT_STANDARD.items():
        if not symbol.startswith('['):
            continue
        if len(rc) and unit.name in rc.Name:
            idx = rc.Name.index(unit.name)
            rc.Symbol[idx] += ", "+symbol
        else:
            rc.append([symbol, unit.name, unit.definition])
    df = rc.to_dataframe()
    file_path = f"{path_docs_static}/constants.csv"
    df.to_csv(file_path, index=False)
    print(file_path)
    
def build_logarithmic_units():
    
    # create table of logarithmic units
    rc = RowCollector(['Symbol','Name','Prefixes'])
    for symbol, unit in UNIT_STANDARD.items():
        if symbol not in ['AR','PR'] and unit.definition != LogarithmicUnitType:
            continue
        if unit.prefixes is True:
            prefixes = 'all'
        elif isinstance(unit.prefixes, list):
            prefixes = ", ".join([f"{p}{symbol}" for p in unit.prefixes])
        else:
            prefixes = ''
        if len(rc) and unit.name in rc.Name:
            idx = rc.Name.index(unit.name)
            rc.Symbol[idx] += ", "+symbol
            rc.Prefixes[idx] += ", "+prefixes if rc.Prefixes[idx] else ""
        else:
            rc.append([symbol, unit.name, prefixes])
    df = rc.to_dataframe()
    file_path = f"{path_docs_static}/unit_logarithmic.csv"
    df.to_csv(file_path, index=False)
    print(file_path)
    
def build_temperature_units():
    
    # create table of temperature units
    rc = RowCollector(['Symbol','Name','Prefixes'])
    for symbol, unit in UNIT_STANDARD.items():
        if symbol not in ['K','degR'] and unit.definition != TemperatureUnitType:
            continue
        if unit.prefixes is True:
            prefixes = 'all'
        elif isinstance(unit.prefixes, list):
            prefixes = ", ".join([f"{p}{symbol}" for p in unit.prefixes])
        else:
            prefixes = ''
        rc.append([symbol, unit.name, prefixes])
    df = rc.to_dataframe()
    file_path = f"{path_docs_static}/unit_temperature.csv"
    df.to_csv(file_path, index=False)
    print(file_path)
    
def build_si_quantities():

    # create table of physical quantities SI
    rc = RowCollector(['Symbol','Name','Definition','Dimensions'])
    for q in range(len(QUANTITY_LIST)):
        name = QUANTITY_LIST.name[q]
        symbol = QUANTITY_LIST.symbol[q]
        definition = QUANTITY_LIST.SI[q]
        magnitude, dimensions = QUANTITY_UNITS[f"#S{symbol}"]
        rc.append([f"#S{symbol}", f"SI.{name}", definition, dimensions])
    df = rc.to_dataframe()
    file_path = f"{path_docs_static}/quantities_si.csv"
    df.to_csv(file_path, index=False)
    print(file_path)
    
def build_cgs_quantities():
    
    # create table of physical quantities CGS
    rc = RowCollector(['Symbol','Name','Definition'])
    for q in range(len(QUANTITY_LIST)):
        symbol = QUANTITY_LIST.symbol[q]
        if f"#C{symbol}" not in QUANTITY_UNITS:
            continue
        name = QUANTITY_LIST.name[q]
        definition = QUANTITY_LIST.CGS[q]
        rc.append([f"#C{symbol}", f"CGS.{name}", definition])
    df = rc.to_dataframe()
    file_path = f"{path_docs_static}/quantities_cgs.csv"
    df.to_csv(file_path, index=False)
    print(file_path)
    
def build_au_quantities():
    
    # create table of physical quantities AU
    rc = RowCollector(['Symbol','Name','Definition'])
    for q in range(len(QUANTITY_LIST)):
        symbol = QUANTITY_LIST.symbol[q]
        if f"#A{symbol}" not in QUANTITY_UNITS:
            continue
        name = QUANTITY_LIST.name[q]
        definition = QUANTITY_LIST.AU[q]
        rc.append([f"#A{symbol}", f"AU.{name}", definition])
    df = rc.to_dataframe()
    file_path = f"{path_docs_static}/quantities_au.csv"
    df.to_csv(file_path, index=False)
    print(file_path)
    
def build_constants():
    
    pt = ParameterTable(PT_HEADER, PT_DATA, keys=True)
    # create table of constants
    rc = RowCollector(['Symbol','Z','A','Relative atomic mass (Da)', 'Natural abundance'])
    for symbol, e in pt.items():
        first = True
        for A, (M, NA) in e.A.items():
            if first:
                rc.append([f"``{symbol}``", e.Z, A, M, NA])
                first = False
            else:
                rc.append(['', '', A, M, NA])
    df = rc.to_dataframe()
    file_path = f"{path_docs_static}/elements.csv"
    df.to_csv(file_path, index=False)
    print(file_path)
    


import numpy as np
import pytest
from math import isclose
from io import StringIO 
import os
import sys
sys.path.insert(0, 'src')

from scinumtools.units import Quantity, Unit 
from scinumtools.materials import *

class Capturing(list):
    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        return self
    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        del self._stringio    # free up some memory
        sys.stdout = self._stdout
        
def test_expression():
    
    # Natural abundance
    e = Element('B')
    assert str(e.mass) == "Quantity(1.081e+01 Da)"
    assert [e.Z,e.N,e.e] == [5.0, 5.801, 5.0]
    assert isclose(e.mass.value(), 10.811028046410001, rel_tol=MAGNITUDE_PRECISION)
    
    # Isotope
    e = Element('B{11}')
    assert str([e.mass,e.Z,e.N,e.e]) == "[Quantity(1.101e+01 Da), 5, 6, 5]"
    assert isclose(e.mass.value(), 11.009306, rel_tol=MAGNITUDE_PRECISION)

    # Isotope and ionisation
    e = Element('B{11-1}')
    assert str([e.mass,e.Z,e.N,e.e]) == "[Quantity(1.101e+01 Da), 5, 6, 4]"
    assert isclose(e.mass.value(), 11.008757420144217, rel_tol=MAGNITUDE_PRECISION)

    # Most abundant
    e = Element('B', natural=False)
    assert str([e.mass,e.Z,e.N,e.e]) == "[Quantity(1.101e+01 Da), 5, 6, 5]"
    assert isclose(e.mass.value(), 11.009306, rel_tol=MAGNITUDE_PRECISION)
    

def test_string():
    
    e = Element('O')
    assert str(e) == "Element(O mass=15.999 Z=8.0 N=8.004 e=8.0)"
    assert str(e.mass) == "Quantity(1.600e+01 Da)"
    assert [e.Z, e.N, e.e] == [8.0, 8.00448, 8.0]

    e = Element('O', natural=False)
    assert str(e) == "Element(O mass=15.995 Z=8 N=8.000 e=8)"
    
    e = Element('O', proportion=2, natural=False)
    assert str(e) == "Element(O2 mass=31.990 Z=16 N=16.000 e=16)"
    assert e.proportion == 2
    assert e.element == 'O'
    assert e.isotope == 16
    assert e.ionisation == 0
    
    e = Element('O{17-2}')
    assert str(e) == "Element(O{17-2} mass=16.998 Z=8 N=9.000 e=6)"

def test_arithmetic():
    
    e1 = Element('B')
    e2 = Element('B',2)

    # addition
    e12 = e1 + e2
    assert e12.proportion == e1.proportion + e2.proportion

    # multiplication
    e13 = e1 * 3
    assert e13.proportion == e1.proportion * 3

def test_matter():
    
    e = Element('B', mass_density=Quantity(997,'kg/m3'), volume=Quantity(1,'l'))
    assert str(e) == "Element(B mass=10.811 Z=5.0 N=5.801 e=5.0)"
    with Capturing() as output:
        e.print()
    assert output == [
        'Element:', 
        '', 
        'Expression: B', 
        'Mass:       10.811', 
        'Protons:    5.0', 
        'Neutrons:   5.801', 
        'Electrons:  5.0', 
        '', 
        'Matter:',
        '',
        'Mass density:   Quantity(9.970e-01 g*cm-3)', 
        'Number density: Quantity(5.554e+22 cm-3)', 
        'Volume:         Quantity(1.000e+00 l)', 
        'Mass:           Quantity(9.970e+02 g)', 
        '', 
        'expr      n[cm-3]  rho[g/cm3]            N  M[g]', 
        '   B 5.553657e+22       0.997 5.553657e+25 997.0'
    ]
    with Capturing() as output:
        e.print_matter()
    assert output == [
        'expr      n[cm-3]  rho[g/cm3]            N  M[g]', 
        '   B 5.553657e+22       0.997 5.553657e+25 997.0'
    ]
    data = e.data_matter()
    assert str(data.B.n) == "Quantity(5.554e+22 cm-3)"
    



from typing import Union, List
import numpy as np

from .type_number import NumberType
from .type_boolean import BooleanType

class FloatType(NumberType):
    value: Union[float,list]
    precision: int = 64
    dtype = float
        
    def __str__(self):
        if self.unit:
            return f"FloatType({self.value} {self.unit})"
        else:
            return f"FloatType({self.value})"
            
    def __repr__(self):
        if self.unit:
            return f"FloatType({self.value} {self.unit})"
        else:
            return f"FloatType({self.value})"
            
    def __init__(self, value, unit=None, **kwargs):
        if isinstance(value, FloatType):
            kwargs['value'] = value.value
            kwargs['unit'] = value.unit
        else:
            kwargs['value'] = value
            kwargs['unit'] = unit
        if isinstance(kwargs['value'], np.ndarray):
            kwargs['value'] = kwargs['value'].tolist()
        if 'precision' in kwargs:
            self.precision = int(kwargs['precision'])
            del kwargs['precision']
        super().__init__(**kwargs)



from bs4 import BeautifulSoup

NTYPES = [
    ('Declaration',                '#698D3F',),
    ('Definition',                 '#AABA78',),
    ('Declaration / Modification', '#D77F33',),
    ('Definition / Modification',  '#DDA15E',),
    ('Modification',               '#FBE974',),
    ('Injection',                  '#AB88BF',),
    ('Import',                     '#D0BCDC',),
]

PAGE_TYPES      = "parameters.html"
PAGE_PARAMETERS = "parameters.html"
PAGE_NODES      = "parameters.html"
PAGE_INJECTIONS = "references.html"
PAGE_IMPORTS    = "references.html"
PAGE_UNITS      = "settings.html"
PAGE_SOURCES    = "settings.html"

def Title(name:str='', h:int=2):
    return f"<a name=\"SECTION_{name}\"></a><h{h}>{name}</h{h}>"

def Target(key:str, name:str=''):
    return BeautifulSoup(f"<a name=\"{key}\"></a>{name}", 'html.parser')
    
def Link(page:str, key:str, name:str=''):
    return BeautifulSoup(f"<a href=\"{page}#{key}\" color=\"blue\">{name}</a>", 'html.parser')


import numpy as np
from typing import List, Dict, Union
from dataclasses import dataclass, field
import copy

from ..units import Quantity, UnitEnvironment
from .datatypes import NumberType
from .settings import *
from .lists import *
    
@dataclass
class Environment:
    # Environment variables lists
    nodes: NodeList          = field(default_factory = NodeList)       # nodes
    units: UnitList          = field(default_factory = UnitList)       # list of cutom units
    sources: SourceList      = field(default_factory = SourceList)     # list of reference sources
    functions: FunctionList  = field(default_factory = FunctionList)   # custom native functions

    # State variable lists
    hierarchy: HierarchyList = field(default_factory = HierarchyList)  # node name hierarchy
    branching: BranchingList = field(default_factory = BranchingList)  # code branching

    # Special mode flags
    autoref: str     = None          # Reference on the current node
    envtype: EnvType = EnvType.DATA  # Documentation mode
    
    def copy(self):
        """ Copy a new object from self
        """
        return copy.deepcopy(self)

    def request(self, path:str, count:int=None, namespace:Namespace=Namespace.NODES, tags:list=None, errsrc:bool=True):
        """ Request nodes from a path

        :param str path: Request path
        :param int count: Number of nodes that should be selected
        :param str namespace: Query namespace (nodes, sources, or units)
        :param list tags: List of tags
        :param bool errsrc: If 'False', missing sources will not throw errors
        """
        if self.autoref and path == Sign.QUERY:     # reference type {?}
            source, query = '', self.autoref
        elif Sign.QUERY in path:                    # reference type {source?query}
            source, query = path.split(Sign.QUERY)
        else:                                       # reference type {source}
            source, query = path,None
        if source:  # use external source to parse the values
            if source not in self.sources and errsrc is False:
                nodes = []
            elif query is None:   # import block
                return self.sources[source].code
            elif namespace == Namespace.NODES:
                nodes = self.sources[source].nodes.query(query, tags=tags)
            elif namespace == Namespace.SOURCES:
                nodes = self.sources[source].sources.query(query)
            elif namespace == Namespace.UNITS:
                nodes = self.sources[source].units.query(query)
            else:
                nodes = self.sources[source].query(query, namespace, tags=tags)
        else:         # use values parsed in the current file
            if not self.nodes:
                raise Exception(f"Local nodes are not available for DIP import:", path)
            nodes = self.nodes.query(query, tags=tags)
        if count:
            if isinstance(count, list) and len(nodes) not in count:
                raise Exception(f"Path returned invalid number of nodes:", path, count, len(nodes))
            elif np.isscalar(count) and len(nodes)!=count:
                raise Exception(f"Path returned invalid number of nodes:", path, count, len(nodes))
        return nodes

    def data(self, format:Format=Format.VALUE, verbose:bool=False, query:str=None, tags:list=None):
        """ Return parsed values as a dictionary

        :param bool verbose: Display node values
        :param str format: Return data as values only, DIP datatypes, or tuples
        :param str query: Node selection query
        :param list tags: List of tags
        """
        data = {}
        if query is not None:
            nodes = self.nodes.query(query, tags=tags)
        elif tags is not None:
            nodes = self.nodes.query("*", tags=tags)
        else:
            nodes = self.nodes
        for node in nodes:
            if format==Format.VALUE:
                data[node.name] = node.value.value
            elif format==Format.TYPE:
                data[node.name] = node.value
            elif format==Format.TUPLE:
                if isinstance(node.value, NumberType) and node.value.unit is not None:
                    data[node.name] = (node.value.value, node.value.unit)
                else:
                    data[node.name] = node.value.value
            elif format==Format.QUANTITY:
                if isinstance(node.value, NumberType):
                    data[node.name] = Quantity(node.value.value, node.value.unit)
                else:
                    data[node.name] = node.value.value
            elif format==Format.NODE:
                data[node.name] = node
            else:
                raise Exception("Data format not recognized:", format)
            if verbose:
                 print(node.name,'|',node.indent,'|',node.keyword,'|',str(node.value),
                       '|',repr(node.units_raw), end='')
                 if hasattr(node,'options'):
                     if node.options:
                         print(' |',node.options, end='') 
                 print()
        return data
        



from .node_base import BaseNode
from ..environment import Environment
from ..settings import Sign, EnvType

class ImportNode(BaseNode):
    keyword: str = 'import'

    def is_node(parser):
        parser.part_reference()
        if parser.is_parsed('part_reference'):
            parser.part_comment()
            return ImportNode(parser)
    
    def inject_value(self, env:Environment, node=None):
        pass
    
    def parse(self, env):
        # Parse import code
        nodes_new = []
        if env.envtype==EnvType.DOCS:
            nodes = env.request(self.value_ref, errsrc=False)
        else:
            nodes = env.request(self.value_ref)
        for node in nodes:
            path = self.name.split(Sign.SEPARATOR + '{')
            path.pop()
            path.append(node.name)     
            node.name = Sign.SEPARATOR.join(path)
            node.indent = self.indent
            node.isource = self.source
            nodes_new.append(node)
        return nodes_new



from bs4 import BeautifulSoup
import json

from .settings import *
from ..item_parameter import ParType
from ..documentation import Documentation

class ParametersSection:
        
    docs: Documentation
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass
    
    def __init__(self, docs: Documentation, **kwargs):
        self.docs = docs
        self.html = BeautifulSoup(f"", 'html.parser')

    def styles(self):
        return None
        
    def build_types(self):
        section = BeautifulSoup(Title("Node types",3), 'html.parser')
        self.html.append(section)
        
        table = self.html.new_tag('table')
        for ptype in ParType:
            row = self.html.new_tag('tr')
            col = self.html.new_tag('td', **{'width': 20, 'style':'background-color:'+NTYPES[ptype.value][1]})
            col.string = ' '
            row.append(col)
            col = self.html.new_tag('td', **{'class':'ps-3'})
            col.string = NTYPES[ptype.value][0]
            row.append(col)
            table.append(row)
            
        self.html.append(table)
        
    def build_parameters(self):
        section = BeautifulSoup(Title("Parameter list",3), 'html.parser')
        self.html.append(section)
            
        table = self.html.new_tag('table', **{'class':'table table-bordered'})
        row = self.html.new_tag('tr', **{'class':'thead-light'})
        col = self.html.new_tag('th', **{'class':'p-1 bg-body-secondary'})
        col.string = "Property name"
        row.append(col)
        for ptype in range(len(ParType)):
            col = self.html.new_tag('th', **{'class':'p-1 text-center','style':'background-color:'+NTYPES[ptype][1]})
            col.string = "#"
            row.append(col)
        table.append(row)
        
        pnames = list(self.docs.parameters.keys())
        pnames.sort()
        for pname in pnames:
            pdata = self.docs.parameters[pname]
            row = self.html.new_tag('tr')
            col = self.html.new_tag('td', **{'class':'p-1'})
            col.append(Link(PAGE_PARAMETERS,pdata.target, pname))
            row.append(col)
            for count in pdata.counts:
                col = self.html.new_tag('td', **{'class':'text-center p-1'})
                col.string = str(count) if count else ''
                row.append(col)
            table.append(row)
        self.html.append(table)

    def build_nodes(self):
        section = BeautifulSoup(Title("Parameter nodes",3), 'html.parser')
        self.html.append(section)
        
        def add_property(pname, pvalue):
            props = self.html.new_tag('div', **{'class':'row'})
            name = self.html.new_tag('div', **{'class':'col-md-2 bg-body-secondary'})
            name.string = pname
            props.append(name)
            value = self.html.new_tag('div', **{'class':'col bg-white'})
            value.string = pvalue
            props.append(value)
            return props
        
        pnames = list(self.docs.parameters.keys())
        pnames.sort()
        for pname in pnames:
            pdata = self.docs.parameters[pname]
            param = self.html.new_tag("h6", **{'class':'mt-2'})
            param.append(Target(pdata.target, pname))
            self.html.append(param)
            
            for ndata in pdata.nodes:
                node = self.html.new_tag("div", **{'class':"container mt-2 border"})
                header = self.html.new_tag('div', **{'class':'row bg-dark-subtle'})
                links = self.html.new_tag('div', **{'class':'col'})
                links.append(Target(ndata.target))
                links.append(Link(PAGE_SOURCES,ndata.link_source, f"{ndata.source[0]}:{ndata.source[1]}"))
                if ndata.injection:
                    links.append(" | ")
                    links.append(Link(PAGE_INJECTIONS,ndata.link_injection, "injection"))
                if ndata.imported:
                    links.append(" | ")
                    links.append(Link(PAGE_IMPORTS,ndata.link_import, "import"))
                header.append(links)
                dtype = self.html.new_tag('div', **{'class':'col-sm-3 text-right'})
                if ndata.constant:
                    dtype.string = "constant "+ndata.dtype
                else:
                    dtype.string = ndata.dtype
                header.append(dtype)
                node.append(header)
                
                if ndata.value:
                    node.append(add_property('Value:', ndata.value))
                if ndata.unit:
                    node.append(add_property('Unit:', ndata.unit))
                if ndata.condition:
                    node.append(add_property('Condition:', ndata.condition))
                if ndata.options:
                    node.append(add_property('Options:', ", ".join(ndata.options)))
                if ndata.tags:
                    node.append(add_property('Tags:', ", ".join(ndata.tags)))
                if ndata.dformat:
                    node.append(add_property('Format:', ndata.dformat))
                if ndata.description:
                    node.append(add_property('Description:', ndata.description))

                self.html.append(node)
            

    def build(self):
        
        self.build_types()
        
        self.build_parameters()
        
        self.build_nodes()

        return self.html


import numpy as np
import pytest
import os
import sys
from decimal import Decimal
import numpy as np
sys.path.insert(0, 'src')

from scinumtools.units import *

def test_decimal():
    
    assert str(Quantity(Decimal(3.3239840203948394e-3), 'cm')) == "Quantity(3.324e-3 cm)"
    
    a = Quantity(Decimal(3.3239840203948394e-3), 'cm')
    b = Quantity(Decimal(9.9239000409020932894e6), 'cm')
    c = Quantity(9.9239000409020932894e6, 'cm')

    # basic arithmetics
    assert str(a+b)      == "Quantity(9.924e+6 cm)"
    assert (a+b).value() == Decimal('9923900.044226077073258914050')
    assert str(a+c)      == "Quantity(9.924e+6 cm)"
    assert str(c+a)      == "Quantity(9.924e+6 cm)"

    assert str(a-b)      == "Quantity(-9.924e+6 cm)"
    assert (a-b).value() == Decimal('-9923900.037578109032469235364')

    assert str(a*b)      == "Quantity(3.299e+4 cm2)"
    assert (a*b).value() == Decimal('32986.88515595424986253677220')

    assert str(a/b)      == "Quantity(3.349e-10)"
    assert (a/b).value() == Decimal('3.349473500030020118277236552E-10')

    assert str(a**2)      == "Quantity(1.105e-5 cm2)"
    assert (a**2).value() == Decimal('0.00001104886976784023973209249265')

    # NumPy functions
    assert str(np.sqrt(a*a))    == "Quantity(3.324e-3 cm)"
    assert np.sqrt(a*a).value() == Decimal('0.003323984020394839342810167081')

    # unit conversion
    assert str(a.to('nm')) == "Quantity(3.324e+4 nm)"
    assert a.value()       == Decimal('33239.84020394839204981469679')



from dataclasses import dataclass, field
import pandas as pd
from typing import Union

class ParameterSettings:
    """ ParameterSettings class contain all settings of a parameter
    """
    
    _keys: list
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, tb):
        pass

    def __getitem__(self, key):
        return getattr(self,key)
    
    def __init__(self, settings):
        self._keys = []
        for key,value in settings.items():
            self._keys.append(key)
            setattr(self,key,value)
            
    def _to_string(self):
        settings = []
        for key in self._keys:
            value = str(getattr(self, key))
            settings.append(f"{key}={value}")
        settings = " ".join(settings)
        return f"ParameterSettings({settings})"
    def __str__(self):
        return self._to_string()
    def __repr__(self):
        return self._to_string()
    def keys(self):
        return self._keys
    def items(self):
        settings = []
        for key in self._keys:
            yield key, getattr(self, key)
    def data(self):
        return {k:v for k,v in self.items()}        

@dataclass
class ParameterTable:
    """ ParameterDict class collects parameters in a concise form and create a dictionary
    """
    _settings: list
    _keys: list = None
    _keyname: str = None
    _data: Union[dict,list] = None

    def __init__(self, settings: list, parameters: Union[list,dict]=None, keys: bool=False, keyname: str="#"):
        self._settings = settings
        if keys:
            self._keys = []
            self._keyname = keyname
        self._data = [] if self._keys is None else {}
        if parameters:
            if self._keys is None:
                for values in parameters:
                    self.append(values) 
            else:
                for key,values in parameters.items():
                    self.append(key, values)            

    def __enter__(self):
        return self
    
    def __exit__(self, type, value, tb):
        pass
    
    def __getitem__(self, key):
        if self._keys is None:
            return self._data[key]
        else:
            if isinstance(key,int):
                return self._data[self._keys[key]]
            else:
                return self._data[key]

    def __setitem__(self, key, values):
        if self._keys is None:
            raise Exception("Cannot set item using a parameter key.")
        else:
            self.append(key, values)

    def __getattr__(self, key):
        if self._keys is None:
            raise Exception("Cannot access parameters using a parameter key.")
        else:   
            return self._data[key]

    def __delitem__(self, index):
        if self._keys is None:
            del self._data[index]
        else:
            self._keys.remove(index)
            del self._data[index]
        
    def __len__(self):
        return len(self._data)
        
    def __contains__(self, item):
        if self._keys is None:
            raise Exception("Parameter table does not have string keys")
        else:
            return item in self._keys
            
    def _to_string(self):
        keys = ", ".join(self._keys)
        return f"ParameterSettings({keys})"
    def __str__(self):
        return self._to_string()
    def __repr__(self):
        return self._to_string()
        
    def shape(self):
        return (len(self._data), len(self._settings))
    
    def keys(self):
        if self._keys is None:
            raise Exception("Parameters do not have keys.")
        else:
            return self._keys
        
    def items(self):
        if self._keys is None:
            return [(key,value) for key,value in enumerate(self._data)]
        else:
            return self._data.items()

    def append(self, *args):
        if self._keys is None:
            values = args[0]
            settings = ParameterSettings(dict(zip(self._settings, values)))
            self._data.append( settings )
        else:
            key, values = args
            if key not in self._keys:
                self._keys.append(key)
            settings = ParameterSettings(dict(zip(self._settings, values)))
            self._data[key] = settings
        
    def data(self):
        if self._keys is None:
            return [v.data() for k,v in self.items()]
        else:
            return {k:v.data() for k,v in self.items()}
            
    def to_dataframe(self):
        if self._keys is None:
            df = pd.DataFrame(columns=self._settings)
            for v in self.data():
                df.loc[len(df.index)] = v.values()
        else:
            df = pd.DataFrame(columns=[self._keyname]+self._settings)
            for k,v in self.data().items():
                df.loc[len(df.index)] = [k]+list(v.values())
        return df
        
    def to_text(self, **kwargs):
        return self.to_dataframe().to_string(**kwargs)



import numpy as np
import os
import pytest
import sys
sys.path.insert(0, 'src')

import scinumtools as snt

@pytest.fixture
def mock_data():
    xlen = 10
    ylen = 20
    data = np.linspace(1,xlen*ylen,xlen*ylen).reshape(xlen,ylen) - 10
    return xlen, ylen, data

def test_data_only(mock_data):
    xlen, ylen, data = mock_data

    # Test only data
    with snt.NormalizeData() as n:
        for row in data:
            n.append(row)
        linnorm = n.linnorm()
        lognorm = n.lognorm()
        zranges = n.zranges()

    assert linnorm.vmin == -9.
    assert linnorm.vmax == 190.0
    assert lognorm.vmin == 0
    assert lognorm.vmax == 2.278753600952829
    assert zranges.minpos == 1
    assert zranges.min == -9
    assert zranges.max == 190.0
    
def test_data_xaxis(mock_data):
    xlen, ylen, data = mock_data

    # Test data and x-axis
    with snt.NormalizeData(xaxis=True) as n:
        for r,row in enumerate(data):
            xdata = np.linspace(-r,r,xlen)
            n.append(row, xdata)
        xranges = n.xranges()

    assert xranges.minpos == 0.11111111111111116
    assert xranges.min == -9
    assert xranges.max == 9

def test_data_xyaxis(mock_data):
    xlen, ylen, data = mock_data

    # Test data and both axes
    with snt.NormalizeData(xaxis=True, yaxis=True) as n:
        for r,row in enumerate(data):
            xdata = np.linspace(-r,r,xlen)
            ydata = np.linspace(-r*2,r*2,ylen)
            n.append(row, xdata, ydata)
        xranges = n.xranges()
        yranges = n.yranges()
        extent = n.extent()

    assert xranges.minpos == 0.11111111111111116
    assert xranges.min == -9
    assert xranges.max == 9
    assert yranges.minpos == 0.10526315789473673
    assert yranges.min == -18
    assert yranges.max == 18
    
def test_iterator(mock_data):
    xlen, ylen, data = mock_data

    with snt.NormalizeData(xaxis='lin', yaxis='lin') as n:
        for r,row in enumerate(data):
            xdata = np.linspace(-r,r,xlen)
            ydata = np.linspace(-r*2,r*2,ylen)
            n.append(row, xdata, ydata)

        item, extent = n[0]
        assert extent == (0,0,0,0)

        for i, (item, extent) in enumerate(n.items()):
            assert extent == (-i,i,-i*2,i*2)
            assert np.all(data[i] == item)


from typing import List, Union
from dataclasses import dataclass

from ..datatypes import IntegerType, FloatType, StringType
from ..settings import EnvType

@dataclass
class Option:
    value: Union[IntegerType, FloatType, StringType]
    value_raw: str
    units_raw: str

class SelectNode:
    
    options: List = None

    def __init__(self, *args, **kwargs):
        self.options = []

    def validate_options(self):
        """ Check if current value is in option list
        """
        if self.options:
            for option in self.options:
                if option.value==self.value:
                    return True
            else:
                options = [o.value for o in self.options]
                raise Exception(f"Value '{str(self.value)}' of node '{self.name}' doesn't match with any option:", options)
        else: # if there are no options return true
            return True
    
    def set_option(self, node, env):
        """ Set option using value of a different node
        """
        if self.keyword=='int':
            value = IntegerType(self.cast_value(node.value_raw), node.units_raw)
            if env.envtype == EnvType.DOCS:
                value.convert(self.units_raw, env)
        elif self.keyword=='float':
            value = FloatType(self.cast_value(node.value_raw), node.units_raw)
            if not env.envtype == EnvType.DOCS:
                value.convert(self.units_raw, env)
        elif self.keyword=='str':
            value = StringType(self.cast_value(node.value_raw))
        self.options.append(Option(
            value=value,
            value_raw=node.value_raw,
            units_raw=node.units_raw
        ))



import numpy as np
import pytest
from math import isclose
from io import StringIO 
import os
import sys
sys.path.insert(0, 'src')

from scinumtools.units import Quantity, Unit 
from scinumtools.materials import *

class Capturing(list):
    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        return self
    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        del self._stringio    # free up some memory
        sys.stdout = self._stdout

def test_empty():
    material = Material()
    assert material.data_components() == None
    
def test_set_dictionary():
    material = Material({
        'H2O': 1.0
    })
    assert material.data_components(quantity=False).to_text() == """
  expr  fraction       mass     Z        N     e
0  H2O       1.0  18.015286  10.0  8.00471  10.0
""".strip('\n')

    material = Material({
        'H2O':  0.2,
        'NaCl': 0.8,
    })
    assert material.data_components(quantity=False).to_text() == """
   expr  fraction       mass     Z         N     e
0   H2O       0.2  18.015286  10.0   8.00471  10.0
1  NaCl       0.8  58.442707  28.0  30.48480  28.0
""".strip('\n')

def test_quantities():
    material = Material({
        'H2O': 1.0
    })
    assert material.data_components(quantity=False).to_text() == """
  expr  fraction       mass     Z        N     e
0  H2O       1.0  18.015286  10.0  8.00471  10.0
""".strip('\n')
    data = material.data_components()
    assert str(data.H2O.mass) == "Quantity(1.802e+01 Da)"
    assert str(data.H2O.Z) == "10.0"
    
def test_expr():
    material = Material('0.2 <H2O> 0.8 <NaCl>')
    assert material.data_components(quantity=False).to_text() == """
   expr  fraction       mass     Z         N     e
0   H2O       0.2  18.015286  10.0   8.00471  10.0
1  NaCl       0.8  58.442707  28.0  30.48480  28.0
""".strip('\n')

def test_print_data():
    material = Material('0.2 <H2O> 0.8 <NaCl>')
    with Capturing() as output:
        material.print_components()
    assert output == [
        'expr  fraction  mass[Da]    Z        N    e', 
        ' H2O       0.2 18.015286 10.0  8.00471 10.0', 
        'NaCl       0.8 58.442707 28.0 30.48480 28.0', 
    ]
    with Capturing() as output:
        material.print_composite()
    assert output == [
        'expr  x[%]       X[%]', 
        ' H2O  20.0   7.154996', 
        'NaCl  80.0  92.845004', 
        ' avg  50.0  50.000000', 
        ' sum 100.0 100.000000',
    ]

def test_norm_number():
    material = Material('2 <H2O> 3 <NaCl>')
    assert material.data_composite(quantity=False).to_text() == """
   expr      x           X
0   H2O   40.0   17.047121
1  NaCl   60.0   82.952879
2   avg   50.0   50.000000
3   sum  100.0  100.000000
""".strip('\n')

def test_norm_number_fraction():
    material = Material('0.2 <H2O> 0.3 <NaCl>')
    assert material.data_composite(quantity=False).to_text() == """
   expr      x           X
0   H2O   40.0   17.047121
1  NaCl   60.0   82.952879
2   avg   50.0   50.000000
3   sum  100.0  100.000000
""".strip('\n')
    
def test_norm_mass_fraction():
    # setting mass fraction
    material = Material('0.2 <H2O> 0.3 <NaCl>', norm_type=Norm.MASS_FRACTION)
    assert material.data_composite(quantity=False).to_text() == """
   expr           x      X
0   H2O   68.381526   40.0
1  NaCl   31.618474   60.0
2   avg   50.000000   50.0
3   sum  100.000000  100.0
""".strip('\n')
    
    # checking if number fractions given above produce the same mass fractions
    material = Material('0.683815 <H2O> 0.316185 <NaCl>', norm_type=Norm.NUMBER_FRACTION)
    assert material.data_composite(quantity=False).to_text() == """
   expr         x           X
0   H2O   68.3815   39.999971
1  NaCl   31.6185   60.000029
2   avg   50.0000   50.000000
3   sum  100.0000  100.000000
""".strip('\n')

def test_natural_abundances():
    # natural abundances
    material = Material({
        'H2O': 1.0
    }, natural=True)
    assert isclose(
        material.data_components(quantity=False).H2O.mass, 
        18.015286, 
        rel_tol=MAGNITUDE_PRECISION
    )
    
    # highest abundances
    material = Material({
        'H2O': 1.0
    }, natural=False)
    assert isclose(
        material.data_components(quantity=False).H2O.mass, 
        18.010565, 
        rel_tol=MAGNITUDE_PRECISION
    )

def test_example():
    m = Material({
       'N2':  78.0840,
       'O2':  20.9460,
       'Ar':  0.93400,
       'CO2': 0.03600,
    }, norm_type=Norm.NUMBER_FRACTION)
    assert m.data_components(quantity=False).to_text() == """
  expr  fraction       mass     Z          N     e
0   N2    78.084  28.013406  14.0  14.007280  14.0
1   O2    20.946  31.998810  16.0  16.008960  16.0
2   Ar     0.934  39.947799  18.0  21.985398  18.0
3  CO2     0.036  44.009546  22.0  22.019660  22.0
""".strip('\n')
    assert m.data_composite(quantity=False).to_text() == """
  expr        x           X
0   N2   78.084   75.517607
1   O2   20.946   23.139564
2   Ar    0.934    1.288131
3  CO2    0.036    0.054698
4  avg   25.000   25.000000
5  sum  100.000  100.000000
""".strip('\n')

def test_matter():
    material = Material('0.2 <H2O> 0.3 <NaCl>')
    assert material.data_components(quantity=False).to_text() == """
   expr  fraction       mass     Z         N     e
0   H2O       0.2  18.015286  10.0   8.00471  10.0
1  NaCl       0.3  58.442707  28.0  30.48480  28.0
""".strip('\n')
    assert material.data_composite(quantity=False).to_text() == """
   expr      x           X
0   H2O   40.0   17.047121
1  NaCl   60.0   82.952879
2   avg   50.0   50.000000
3   sum  100.0  100.000000
""".strip('\n')
    material = Material('0.2 <H2O> 0.3 <NaCl>', mass_density=Quantity(0.3,'g/cm3'))
    assert material.data_matter(quantity=False).to_text() == """
   expr             n       rho
0   H2O  1.709551e+21  0.051141
1  NaCl  2.564326e+21  0.248859
2   avg  2.136939e+21  0.150000
3   sum  4.273877e+21  0.300000
""".strip('\n')
    material = Material('0.2 <H2O> 0.3 <NaCl>', mass_density=Quantity(0.3,'g/cm3'), volume=Quantity(1,'l'))
    assert material.data_matter(quantity=False).to_text() == """
   expr             n       rho             N           M
0   H2O  1.709551e+21  0.051141  1.709551e+24   51.141364
1  NaCl  2.564326e+21  0.248859  2.564326e+24  248.858636
2   avg  2.136939e+21  0.150000  2.136939e+24  150.000000
3   sum  4.273877e+21  0.300000  4.273877e+24  300.000000
""".strip('\n')

def test_print():
    material = Material('0.2 <H2O> 0.3 <NaCl>', mass_density=Quantity(0.3,'g/cm3'), volume=Quantity(1,'l'))
    with Capturing() as output:
        material.print()
    assert output == [
        'Components:', 
        '', 
        'expr  fraction  mass[Da]    Z        N    e', 
        ' H2O       0.2 18.015286 10.0  8.00471 10.0', 
        'NaCl       0.3 58.442707 28.0 30.48480 28.0', 
        '', 
        'Composite:', 
        '', 
        'expr  x[%]       X[%]', 
        ' H2O  40.0  17.047121', 
        'NaCl  60.0  82.952879', 
        ' avg  50.0  50.000000', 
        ' sum 100.0 100.000000', 
        '',
        'Matter:', 
        '', 
        'Mass density:   Quantity(3.000e-01 g*cm-3)', 
        'Number density: Quantity(8.548e+21 cm-3)', 
        'Volume:         Quantity(1.000e+00 l)', 
        'Mass:           Quantity(3.000e+02 g)', 
        '', 
        'expr      n[cm-3]  rho[g/cm3]            N       M[g]', 
        ' H2O 1.709551e+21    0.051141 1.709551e+24  51.141364', 
        'NaCl 2.564326e+21    0.248859 2.564326e+24 248.858636', 
        ' avg 2.136939e+21    0.150000 2.136939e+24 150.000000', 
        ' sum 4.273877e+21    0.300000 4.273877e+24 300.000000'
    ]


from .quantity import Quantity
from .quantity  import Quantity as quant
from .unit import Unit
from .unit import Unit as unit
from .constant import Constant
from .constant import Constant as const
from .nan import NaN
from .nan import NaN as nan
from .magnitude import Magnitude
from .dimensions import Dimensions
from .fraction import Fraction
from .base_units import BaseUnits
from .unit_solver import UnitSolver
from .unit_environment import UnitEnvironment
from .systems import SI, CGS, AU



NTYPES = [
    ('Declaration',                '#698D3F',),
    ('Definition',                 '#AABA78',),
    ('Declaration / Modification', '#D77F33',),
    ('Definition / Modification',  '#DDA15E',),
    ('Modification',               '#FBE974',),
    ('Injection',                  '#AB88BF',),
    ('Import',                     '#D0BCDC',),
]

PAGE_TYPES      = "parameters.html"
PAGE_PARAMETERS = "parameters.html"
PAGE_NODES      = "parameters.html"
PAGE_INJECTIONS = "references.html"
PAGE_IMPORTS    = "references.html"
PAGE_UNITS      = "settings.html"
PAGE_SOURCES    = "settings.html"




from bs4 import BeautifulSoup
import os

from .settings import *
from .section_parameters import ParametersSection
from .section_references import ReferencesSection
from .section_settings import SettingsSection
from ..documentation import Documentation

class ExportDocsHTML:
        
    docs: Documentation

    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass
    
    def __init__(self, docs: Documentation, **kwargs):
        self.docs = docs
        
    def build(self, dir_build: str, title: str, intro=None):

        sections = {
            "Parameters": {
                'filename':'parameters',
                'content':  ParametersSection,
                'sections': {
                    "types":"Node types",
                    "parameters":"Parameter list",
                    "nodes":"Parameter nodes"
                }
            },
            "References": {
                'filename':'references',
                'content':  ReferencesSection,
                'sections': {
                    "injections":"Injections",
                    "imports":"Imports"
                }
            },
            "Settings":   {
                'filename':'settings',
                'content':  SettingsSection,
                'sections': {
                    "units":"Units",
                    "sources":"Sources"
                }
            },
        }        
        
        # parse default layout
        
        file_layout = os.path.dirname(os.path.abspath(__file__))+'/layout.html'
        with open(file_layout, 'r', encoding='utf-8') as file:
            html = BeautifulSoup(file.read(), 'html.parser')
            html.find(id='title_docs').string = title
        
        # parse menu
        
        menu = html.find(id='menu')
        row = BeautifulSoup(f"<div><a href='./index.html'>Index</a></div>", 'html.parser')
        menu.append(row)

        for i2, (name, settings) in enumerate(sections.items()):
            
            row = BeautifulSoup(f"<div><a href='./{settings['filename']}.html#SECTION_{name}'>{name}</a></div>", 'html.parser')
            menu.append(row)
            
            sect = settings['sections'].items() if settings['sections'] else []
            for i3, (target,name3) in enumerate(sect):
                row = BeautifulSoup(f"<div class='row ps-4'><a href='./{settings['filename']}.html#SECTION_{name3}'>{name3}</a></div>", 'html.parser')
                menu.append(row)
    
        # parse index
    
        title = html.find(id='title_section')
        content = html.find(id='content')
        style = html.find(id='styles')

        title.append(BeautifulSoup(Title("Index"), 'html.parser')) 
        if intro:
            title.append(BeautifulSoup(Title('Introduction',3), 'html.parser'))
            paragraph = html.new_tag("p")
            paragraph.append(BeautifulSoup(intro, 'html.parser'))
            content.append(paragraph)
        
        with open(f"{dir_build}/index.html", "w") as file:
            file.write(str(html.prettify()))
        
        # parse sections
        
        for i2, (name, settings) in enumerate(sections.items()):
            content.clear()
            title.clear()
            style.string = ''
            title.append(BeautifulSoup(Title(name), 'html.parser'))
            with settings['content'](self.docs) as sect:
                if css := sect.styles():
                    style.string = css
                content.append(sect.build())
            with open(f"{dir_build}/{settings['filename']}.html", "w") as file:
                file.write(str(html.prettify()))
        


import os
from pathlib import Path

import scinumtools.dip as dipsl
from .node_base import BaseNode
from .parser import Parser
from ..settings import Namespace

class SourceNode(BaseNode):
    keyword: str = 'source'

    @staticmethod
    def is_node(parser):
        parser.kwd_source()
        if parser.is_parsed('kwd_source'):
            parser.part_comment()
            return SourceNode(parser)
            
    def parse(self, env):
        parser = Parser(
            code=self.value_raw,
            source=self.source
        )
        # import a remote source
        parser.part_reference()
        if parser.is_parsed('part_reference'):
            sources = env.request(parser.value_ref, namespace=Namespace.SOURCES)
            # append remote sources to the local sources
            for key, val in sources.items():
                env.sources[key] = val
        else:
            # inject value of a node
            parser.part_name(path=False) # parse name
            parser.part_equal()          # parse equal sign
            parser.part_value()          # parse value
            if parser.value_ref:
                self.inject_value(env, parser)
            if parser.value_raw.endswith('dip'):
                # source is a DIP file
                p = dipsl.DIP(source=self.source)
                p.env.sources = env.sources.copy()
                p.add_file(parser.value_raw, parser.name)
                penv = p.parse()
                env.sources[parser.name] = penv.sources[parser.name]
                env.sources[parser.name].nodes = penv.nodes
                env.sources[parser.name].sources = penv.sources
            else:
                # source is a text file
                filepath = parser.value_raw
                if not os.path.isabs(filepath):
                    # set relative paths with respect to the source script
                    source = env.sources[self.source[0]]
                    parent = Path(source.path).parent
                    filepath = parent / filepath
                with open(filepath,'r') as f:
                    env.sources.append(parser.name, filepath, f.read(), self.source[0])
        return None



import numpy as np

from ...units import Quantity, UnitEnvironment
from ...solver import *
from ..nodes.parser import Parser
from ..environment import Environment

class NumericalSolver:

    env: Environment

    def __init__(self, env:Environment=None, **kwargs):
        if env:
            self.env = env
        else:
            self.env = Environment()
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass

    def _parse_atom(self, expr):
        expr = expr.lstrip()
        kwargs = {'code': expr, 'source':None}
        p = Parser(keyword='node',**kwargs)
        p.part_value()
        if p.value_ref:   # import existing node
            nodes = self.env.request(p.value_ref, count=[0,1])
            if len(nodes)==1:
                node = nodes[0]
                value, units = node.value.value, node.value.unit
            else:
                raise Exception('Reference does not return any nodes:', p.value_ref)
        else:             # create anonymous node
            p.part_units()
            value, units = p.value_raw, p.units_raw
        with UnitEnvironment(self.env.units):                
            unit = Quantity(float(value), units)
        unit.symbol = expr
        return unit
        
    def solve(self, expr, in_units=None):
        expr_bak = expr
        # immediately return boolean values
        if isinstance(expr,(int,float)):
            return expr
            
        # solve expression
        with UnitEnvironment(self.env.units):
            operators = {
                'log':CustomOperatorLog, 'log10':CustomOperatorLog10, 'logb':CustomOperatorLogb,
                'exp':CustomOperatorExp, 'sqrt':CustomOperatorSqrt,   'powb':CustomOperatorPowb,
                'sin':CustomOperatorSin, 'cos':CustomOperatorCos,     'tan':CustomOperatorTan,
                'par':OperatorPar,  # should be the last of parenthesis operators
                'pow':OperatorPow,
                'mul':CustomOperatorMul, 'truediv':CustomOperatorTruediv,
                'add':CustomOperatorAdd, 'sub':CustomOperatorSub,
            }
            with ExpressionSolver(self._parse_atom, operators) as es:
                result = es.solve(expr)
            if in_units:
                return result.value(in_units)
            else:
                return result
            
    def equal(self, expr1, expr2):
        unit1 = self.solve(expr1)
        unit2 = self.solve(expr2)
        if not unit1.baseunits.nodim:
            unit2.to(unit1.baseunits)
        if unit1 == unit2:
            return True
        else: 
            return False

class CustomOperatorAdd(OperatorAdd):
    symbol: str = ' + '
    def operate_unary(self, tokens):
        left, right = tokens.get_left(), tokens.get_right()
        if left is None and isinstance(right, Quantity):
            tokens.put_left(right)        
        elif isinstance(right, CustomOperatorAdd):
            tokens.put_left(left)
            tokens.put_right(right)
        elif isinstance(right, CustomOperatorSub):
            tokens.put_left(left)
            tokens.put_right(right)
        elif not isinstance(left, Quantity) and isinstance(right, Quantity):
            tokens.put_left(left)
            tokens.put_right(right)
        else:
            tokens.put_left(left)
            tokens.put_left(CustomOperatorAdd())
            tokens.put_right(right)
    def operate_binary(self, tokens):
        left, right = tokens.get_left(), tokens.get_right()
        if not left.baseunits.nodim:
            right.to(left.baseunits)
        tokens.put_left(left + right)

class CustomOperatorSub(OperatorSub):
    symbol: str = ' - '
    def operate_unary(self, tokens):
        left, right = tokens.get_left(), tokens.get_right()
        if left is None and isinstance(right, Quantity):
            tokens.put_left(-right)
        elif isinstance(right, CustomOperatorAdd):
            tokens.put_left(left)
            tokens.put_right(CustomOperatorSub())
        elif isinstance(right, CustomOperatorSub):
            tokens.put_left(left)
            tokens.put_right(CustomOperatorAdd())
        elif not isinstance(left, Quantity) and isinstance(right, Quantity):
            tokens.put_left(left)
            tokens.put_right(-right)
        else:
            tokens.put_left(left)
            tokens.put_left(CustomOperatorSub())
            tokens.put_right(right)
    def operate_binary(self, tokens):
        left, right = tokens.get_left(), tokens.get_right()
        if not left.baseunits.nodim:
            right.to(left.baseunits)
        tokens.put_left(left - right)

class CustomOperatorMul(OperatorMul):
    symbol: str = ' * '

class CustomOperatorTruediv(OperatorTruediv):
    symbol: str = ' / '

class CustomOperatorExp(OperatorExp):
    def operate_args(self, tokens):
        tokens.put_left(Quantity(np.e)**self.args[0].value())        
            
class CustomOperatorLog(OperatorLog):
    def operate_args(self, tokens):
        tokens.put_left(np.log(self.args[0]))     

class CustomOperatorLog10(OperatorLog10):
    def operate_args(self, tokens):
        tokens.put_left(np.log10(self.args[0]))        

class CustomOperatorSqrt(OperatorSqrt):
    def operate_args(self, tokens):
        tokens.put_left(np.sqrt(self.args[0]))      

class CustomOperatorSin(OperatorSin):
    def operate_args(self, tokens):
        tokens.put_left(np.sin(self.args[0]))      

class CustomOperatorCos(OperatorCos):
    def operate_args(self, tokens):
        tokens.put_left(np.cos(self.args[0]))       

class CustomOperatorTan(OperatorTan):
    def operate_args(self, tokens):
        tokens.put_left(np.tan(self.args[0]))       

class CustomOperatorLogb(OperatorLogb):
    def operate_args(self, tokens):
        tokens.put_left(np.log(self.args[0])/np.log(self.args[1]))

class CustomOperatorPowb(OperatorPowb):
    def operate_args(self, tokens):
        tokens.put_left(np.power(self.args[0],self.args[1].value()))     




import re

from .node_base import BaseNode
from ..settings import Sign, Keyword
from ..solvers import LogicalSolver
from ..datatypes import BooleanType

class CaseNode(BaseNode):
    keyword: str = 'case'
    case_id: int = 0
    case_type: str = None

    @staticmethod
    def is_node(parser):
        parser.kwd_case()
        if parser.is_parsed('kwd_case'):
            if parser.name.endswith(Sign.CONDITION + Keyword.CASE):
                parser.part_value()
            parser.part_comment()
            return CaseNode(parser)
            
    def parse(self, env):
        if m := re.match(fr"(.*{Sign.CONDITION})({Keyword.CASE}|{Keyword.ELSE}|{Keyword.END})$", self.name):
            self.case_id = env.branching.register_case()  # set node case ID
            self.case_type = m.group(2)
            self.name = f"{m.group(1)}{str(self.case_id)}" 
            if m.group(2) == Keyword.CASE:
                with LogicalSolver(env) as s:
                    if self.value_expr:
                        self.value = s.solve(self.value_expr)
                    else:
                        self.value = s.solve(self.value_raw)
            elif m.group(2) == Keyword.ELSE:
                self.value = BooleanType(True)
        return None



from typing import Union, List
import numpy as np

from .type import Type
from .type_boolean import BooleanType

class StringType(Type):    
    value: Union[str,list]
    dtype = str

    def __str__(self):
        return f"StringType({self.value})"
    
    def __repr__(self):
        return f"StringType({self.value})"
        
    def __eq__(self, other):
        return self.value == other.value
        
    def __init__(self, value, **kwargs):
        kwargs['value'] = value
        if isinstance(kwargs['value'], np.ndarray):
            kwargs['value'] = kwargs['value'].tolist()
        super().__init__(**kwargs)



from reportlab.platypus import Paragraph, Table, Spacer, PageBreak
from reportlab.lib.units import inch
import numpy as np
import re

from .settings import *
from ...nodes import ModNode

class InjectionsSection:
    
    nodes: list      # node list
    
    def __enter__(self):
        return self
    def __exit__(self, type, value, traceback):
        pass
        
    def __init__(self, data):
        self.data = data
                           
    def parse_item(self, item):
    
        # define table style
        TABLE_STYLE = [
            # whole grid
            ('GRID',       (0,1), (-1,-1),  0.5, PALETTE['prop_name']),
            ('FONTNAME',   (0,0), (-1,-1),  TABLE_BODY_STYLE.fontName),
            ('FONTSIZE',   (0,0), (-1,-1),  TABLE_BODY_STYLE.fontSize),
            # top panel
            ('FONTSIZE',   (0,0), (-1,-1),  TABLE_HEADER_STYLE.fontSize),
            ('TEXTCOLOR',  (0,0), (-1,0),   PALETTE['c4']),   
            ('BACKGROUND', (0,0), (-1,0),   PALETTE['node_name']),
            ('ALIGN',      (-2,0),(-1,0),   'RIGHT'),
            ('BACKGROUND', (0,1), (0,-1),   PALETTE['prop_name']),  
            ('BACKGROUND', (1,1), (1,-1),  PALETTE['prop_value']),  
        ]
        
        # construct a item table
        target_injection = Target(item.target)
        link_source = Link(item.link_source, f"{item.source[0]}:{item.source[1]}") 
        data = [
            [Paragraph(target_injection+link_source), ''],
        ]
        data.append(['Injecting node:', Paragraph(Link(item.link_node, item.name),          TABLE_BODY_STYLE)])
        data.append(['Request:',        Paragraph(HighlightReference("{"+item.reference+"}"), TABLE_BODY_STYLE)])
        if item.isource:
            data.append(['From source:', Paragraph(Link(item.link_isource, f"{item.isource[0]}:{item.isource[1]}"), TABLE_BODY_STYLE)])
        if item.ivalue:
            data.append(['Value:', Paragraph(item.ivalue, TABLE_BODY_STYLE)])
        if item.iunit:
            data.append(['Unit:',  Paragraph(item.iunit, TABLE_BODY_STYLE)])

        return Table(data,style=TABLE_STYLE, hAlign='LEFT', colWidths=ColumnWidths([0.2, 0.8]))
        
    def parse(self):
        blocks = []
        blocks.append(Paragraph(Title(f"Injected values"), H2) )
        for item in self.data:
            blocks.append(Spacer(1,0.1*inch))
            blocks.append(self.parse_item(item))
        blocks.append(Spacer(1,0.2*inch))
        return blocks
        


from .settings import *
from ..nodes import ModNode, ImportNode, StringNode
from ..settings import Sign, Keyword, EnvType

class InjectionItem:
    target: str
    link_source: str
    link_node: str
    link_isource: str
    name: str               
    reference: str          
    source: tuple           
    isource: tuple = None   
    ivalue: str = None
    iunit: str = None
    
    def __init__(self, node, env):
        self.target = InjectionItemTarget(node.name, node.source[0], node.source[1])
        self.link_source = SourceItemTarget(node.source[0], node.source[1])
        self.link_node = NodeItemTarget(node.name, node.source[0], node.source[1])
        self.name = node.name
        self.reference = node.value_ref
        self.source = node.source
        ref_source, ref_node = node.value_ref.split(Sign.QUERY)
        if ref_source in env.sources:
            inode = env.request(node.value_ref)[0]
            if inode.keyword==ModNode.keyword:
                self.ivalue = inode.value
                if inode.units_raw:
                    self.iunit = inode.units_raw
            else:
                if inode.value.value:
                    self.ivalue = str(inode.value.value)
                if inode.value.unit:
                    self.iunit = inode.value.unit
            self.isource = inode.source
            self.link_isource = SourceItemTarget(inode.source[0], inode.source[1])



from pygments.lexer import RegexLexer, include, bygroups
from pygments.token import Token
    
class SchemaLexer(RegexLexer):
    name = 'DIP'
    tokens = {
        'root': [
            (r'<indent>',     Token.DIP.Indent),
            (r'<name>',       Token.DIP.Name),
            (r'<type>',       Token.DIP.Type),
            (r'<unit>',       Token.DIP.Unit),
            (r'<value>',      Token.DIP.String),
            (r'<path>',       Token.DIP.String),
            (r'\.\.\.',       Token.DIP.Text),
            (r'(\()("""\n|"|\')(<expression>)(\n"""|"|\')(\))',   # """
              bygroups(Token.DIP.Text, Token.DIP.Expression, Token.DIP.Expression,
                       Token.DIP.Expression, Token.DIP.Text)),
            (r'(\()(<function>)(\))',
              bygroups(Token.DIP.Text, Token.DIP.Expression,  Token.DIP.Text)),
            (r'<[^ ]',        Token.DIP.Tag, 'tag'),
            (r'{',            Token.DIP.Reference, 'reference'),
            (r'!(options|format|condition|constant|tags|description)',  Token.DIP.Keyword),
            (r'\$(source|unit)',     Token.DIP.Keyword),
            (r'\@(case|else|end)',     Token.DIP.Keyword),
            (r'( |=|<|>|\.\*|\*)',    Token.DIP.Text),
            (r'#[^\n]*',  Token.DIP.Comment),
            (r'[ ]*\n',   Token.DIP.Text), 
        ],
        'tag': [
            (r'[a-zA-Z0-9_-]+', Token.DIP.Tag),
            (r'>',              Token.DIP.Tag, '#pop'),
        ],
        'reference': [
            (r'[^}]+',          Token.DIP.Reference),
            (r'{',              Token.DIP.Reference, '#push'),
            (r'}\[',            Token.DIP.Reference, 'reference_slice'),
            (r'}',              Token.DIP.Reference, '#pop'),
        ],
        'reference_slice': [
            (r'[^\]]+',         Token.DIP.Slice),
            (r',',              Token.DIP.Reference),
            (r'\]',             Token.DIP.Reference, '#pop:2'),
        ],

    }



import numpy as np

from .export import ExportConfig
from ..datatypes import StringType, BooleanType, NumberType, FloatType, IntegerType
from ..settings import Sign, Format
from ..environment import Environment

class ExportConfigBash(ExportConfig):
    
    def __init__(self, env: Environment, **kwargs):
        if 'dtype' not in kwargs:
            kwargs['dtype'] = Format.VALUE
        super().__init__(env, **kwargs)
        
    def _parse_scalar(self, value):
        if value is None:
            value = ''
        elif isinstance(value, str):
            value = f"\"{value}\""
        elif isinstance(value, bool):
            value = "0" if value else "-1"   # in bash 0 is true and usually 1, -1 for error
        return value
    
    def _parse_array(self, name, values, coord):
        strings = []
        for v, value in enumerate(values):
            if isinstance(value,(np.ndarray,tuple,list)):
                string, shape = self._parse_array(name, value, coord+[v])
            else:
                string, shape = self._parse_scalar(value), None
            strings.append( string )
        if shape is None:
            shape = [len(values)]
            if len(coord):
                dims = ",".join(str(d) for d in coord)
                strings = [f"{name}[{dims},{v}]={val}" for v,val in enumerate(strings)]
        else:
            shape = [len(values)]+shape
        if len(coord)==0 and len(shape)==1:
            return "(" + " ".join(f"\"{val}\"" for val in strings) + ")", shape
        else:
            return "\n".join(str(val) for val in strings), shape
    
    def parse(self, export=True):
        lines = []
        for name, value in self.data.items():
            name = self._rename(name)
            
            if isinstance(value, (list, tuple, np.ndarray)):
                value, size = self._parse_array(name, value, [])
            else:
                value, size = self._parse_scalar(value), None
            export = 'export ' if export else ''
            if size and len(size)>1:
                lines.append(f"declare -A {name}\n{value}")
                if export:
                    lines.append(f"export {name}")
            else:
                lines.append(f"{export}{name}={value}")
        self.text = Sign.NEWLINE.join(lines)
        return self.text


from .node_base import BaseNode

class ModNode(BaseNode):
    keyword: str = 'mod'

    @staticmethod
    def is_node(parser):         
        parser.part_equal()  
        if parser.is_parsed('part_equal'): 
            parser.part_value()
            parser.part_units()
            parser.part_comment()
            return ModNode(parser)
            
    def parse(self, env):
        return None    



from dataclasses import dataclass, field
from typing import List, Dict, Union
import numpy as np

from ..settings import Order, Sign

@dataclass
class NodeList:
    nodes: List     = field(default_factory = list)  # list of nodes
    
    def __len__(self):
        return len(self.nodes)
    
    def __delitem__(self, key):
        del self.nodes[key]
    
    def __getitem__(self, key: Union[int, str]):
        if isinstance(key, int):
            return self.nodes[key]
        elif isinstance(key, str):
            nodes = NodeList()
            for node in self.nodes:
                if node.name == key:
                    return node.copy()
                elif node.name.startswith(key+Sign.SEPARATOR):
                    node = node.copy()
                    node.name = node.name[len(key)+1:]
                    nodes.append(node)
            return nodes
        else:
            raise Exception("Node list keys can be only integers or strings:", key)
        
    def keys(self):
        keys = []
        for node in self.nodes:
            node_key = node.name.split(Sign.SEPARATOR)[0]
            if node_key not in keys:
                keys.append(node_key)
        keys.sort()
        return keys
        
    def pop(self):
        return self.nodes.pop(0)
    
    def append(self, node):
        self.nodes.append(node)
        
    def prepend(self, nodes):
        self.nodes = nodes + self.nodes
        
    def query(self, query:str, tags:list=None, order:Order=None):
        """ Select local nodes according to a query

        :param str query:   Node selection query
        :param list tags:   List of tags
        :param Order order: List ordering
        """
        nodes = NodeList()
        if query==Sign.WILDCARD:
            nodes = NodeList([node.copy() for node in self.nodes])
        elif query[-2:]==Sign.SEPARATOR + Sign.WILDCARD:
            for node in self.nodes:
                if node.name.startswith(query[:-1]):
                    node = node.copy()
                    node.name = node.name[len(query[:-1]):]
                    nodes.append(node.copy())
        else:
            for node in self.nodes:
                if node.name==query:
                    node = node.copy()
                    node.name = node.name.split(Sign.SEPARATOR)[-1]
                    nodes.append(node.copy())
        if tags:
            tagged = NodeList()
            for n in range(len(nodes)):
                if nodes[n].tags and np.isin(tags, nodes[n].tags):
                    tagged.append(nodes[n])
            nodes = tagged
        if order:
            nodes = list(dict(sorted({node.name:node for node in nodes}.items())).values())
        return nodes



import numpy as np
import re
from typing import Union

from . import Norm, Units
from .matter import Matter
from .composite import Composite
from .substance import Substance
from .material_solver import MaterialSolver
from ..units import Quantity, Unit

class Material(Composite, Matter):
    
    def atom(self, expr:str):
        if m:=re.match("[0-9]+(\.[0-9]+|)([eE]{1}[+-]?[0-9]{0,3}|)",expr):
            return float(expr)
        else:
            return Material({expr: 1.0}, natural=self.natural, norm_type=self.norm_type)

    def __init__(self, expr:Union[str,dict]=None, natural:bool=True, norm_type:Norm=Norm.NUMBER_FRACTION, **kwargs):
        Matter.__init__(self, **kwargs)
        Composite.__init__( self,
            MaterialSolver, Substance, expr, {  
                'fraction': None,       
                "mass": Units.ATOMIC_MASS,
                'Z': None,
                'N': None,
                'e': None,
            }, {}, natural=natural, norm_type=norm_type
        )

    def __str__(self):
        substances = []
        for expr, mol in self.components.items():
            substances.append(f"{mol.proportion} {expr}")
        substances = "; ".join(substances)
        return f"Material({substances})"
        
    def __rmul__(self, other:float):
        return self._multiply(Material(natural=self.natural, norm_type=self.norm_type), other)

    def __add__(self, other):
        return self._add(Material(natural=self.natural, norm_type=self.norm_type), other)

    def _add_expr(self, expr:str, proportion:int):
        self.expr += f"{proportion} <{expr}>"
        
    def data_components(self, quantity:bool=True):
        def fn_row(s,m):
            subs = m.data_composite()
            return {
                'fraction': m.proportion,
                'mass':     m.component_mass,
                'Z':        subs.sum.Z,
                'N':        subs.sum.N,
                'e':        subs.sum.e,
            }
        return self._data(self.cols_components, fn_row, quantity=quantity)

    def data_composite(self, components:list=None, quantity:bool=True):
        def fn_row(s,m):
            return {}
        return self._data(self.cols_composite, fn_row, stats=True, weight=False, components=components, quantity=quantity)
            


from . import Units
from ..units import Quantity

class Matter:

    mass_density: Quantity = None   # mass density
    number_density: Quantity = None # number density
    volume: Quantity = None         # volume
    mass: Quantity = None           # mass

    def __init__(self, 
        number_density:Quantity=None, mass_density:Quantity=None, volume:Quantity=None
    ):
        self.cols_matter = {
            "n":     Units.NUMBER_DENSITY, 
            "rho":   Units.MASS_DENSITY,
            "N":     None,
            "M":     Units.MATERIAL_MASS,
        }
        self.number_density = number_density
        self.mass_density = mass_density
        self.volume = volume

    def _norm(self):
      # setup densities of the composite
        if self.mass_density:
            self.mass_density.to(Units.MASS_DENSITY)
            self.number_density = (self.mass_density/self.composite_mass).to(Units.NUMBER_DENSITY)
        elif self.number_density: # !! number density of a composite, not sum of all its components
            self.mass_density = (self.number_density*self.composite_mass).to(Units.MASS_DENSITY)
            self.number_density.to(Units.NUMBER_DENSITY)
        if self.volume:
            self.mass = (self.mass_density * self.volume).to(Units.MATERIAL_MASS)
            
    def _print(self):
        print("Matter:")
        print("")
        if self.mass_density:
            print(f"Mass density:   {self.mass_density}")
            print(f"Number density: {self.number_density}")
            if self.volume:
                print(f"Volume:         {self.volume}")
                print(f"Mass:           {self.mass}")
        print("")
        self._print_table(self.cols_matter, self.data_matter)
        
    def data_matter(self, components:list=None, quantity:bool=True):
        def fn_row(s,m):
            values = {
                'proportion': m.proportion
            }
            if self.number_density:
                values['n']   = m.proportion*self.number_density
                values['rho'] = m.proportion*m.component_mass*self.number_density
            if self.volume:
                values['N'] = values['n']*self.volume
                values['M'] = values['rho']*self.volume
            return values
                # create a new copy of the column list and remove unused columns
        columns = dict(self.cols_matter) 
        if not self.mass_density:
            if 'n' in columns: del columns['n']
            if 'rho' in columns: del columns['rho']
        if not self.volume:
            if 'N' in columns and 'M' in columns: del columns['N']
            if 'M' in columns: del columns['M']
        return self._data(columns, fn_row, stats=True, weight=True, components=components, quantity=quantity)

    def print_matter(self, components:list=None):
        self._print_table(self.cols_matter, self.data_matter, components=components)



import sys, os
import pytest
import numpy as np
sys.path.insert(0, 'src')

from scinumtools.dip import DIP
from scinumtools.dip.config import *

@pytest.fixture
def file_pdf():
    dir_tmp = "tmp" #docs/source/_static/pdf"
    if not os.path.isdir(dir_tmp):
        os.mkdir(dir_tmp)
    file_pdf  = f"{dir_tmp}/export_config.txt"
    if os.path.isfile(file_pdf):
        os.remove(file_pdf)
    return file_pdf

@pytest.fixture
def basic_types():
    return """
    simulation
      name str = 'Configuration test'
      output bool = true
    box
      height float = 15 cm
    num_cells int = 100
      !tags ["selection"]
    """

@pytest.fixture
def derived_types():
    return """
    box
      width float32 = 12 cm
        !tags ["selection"]
    density float128 = 23 g/cm3    
    num_groups uint64 = 2399495729
    """

@pytest.fixture
def arrays():
    return """
    primes int[3] = [3,5,7]
    sizes float[3] = [23.4,46,96.4] cm
    """
    
@pytest.fixture
def none_value():
    return """
    particles
      stars int = none
      tracers int = 23
    """
    
def test_parser_dip(basic_types, derived_types):

    with DIP() as dip:
        dip.add_string(basic_types)
        dip.add_string(derived_types)
        env = dip.parse()
    with ExportConfig(env) as exp:
        assert exp.parse() == """
simulation.name str = "Configuration test"
simulation.output bool = true
box.height float = 15.0 cm
num_cells int = 100
box.width float32 = 12.0 cm
density float128 = 23.0 g/cm3
num_groups uint64 = 2399495729
        """.strip()
        
def test_parser_c(basic_types, derived_types, none_value):

    with DIP() as dip:
        dip.add_string(basic_types)
        dip.add_string(derived_types)
        dip.add_string(none_value)
        env = dip.parse()
    with ExportConfigC(env) as exp:
        assert exp.parse() == """
#ifndef CONFIG_H
#define CONFIG_H

#include <stdbool.h>

const char* SIMULATION_NAME = "Configuration test";
const bool SIMULATION_OUTPUT = true;
const double BOX_HEIGHT = 15.0;
const int NUM_CELLS = 100;
const float BOX_WIDTH = 12.0;
const long double DENSITY = 23.0;
const unsigned long long int NUM_GROUPS = 2399495729;
const int PARTICLES_STARS = None;
const int PARTICLES_TRACERS = 23;

#endif /* CONFIG_H */
        """.strip()

def test_parser_cpp(basic_types, derived_types):

    with DIP() as dip:
        dip.add_string(basic_types)
        dip.add_string(derived_types)
        env = dip.parse()
    with ExportConfigCPP(env) as exp:        
        assert exp.parse() == """
#ifndef CONFIG_H
#define CONFIG_H

constexpr char* SIMULATION_NAME = "Configuration test";
constexpr bool SIMULATION_OUTPUT = true;
constexpr double BOX_HEIGHT = 15.0;
constexpr int NUM_CELLS = 100;
constexpr float BOX_WIDTH = 12.0;
constexpr long double DENSITY = 23.0;
constexpr unsigned long long int NUM_GROUPS = 2399495729;

#endif /* CONFIG_H */
        """.strip()
        
def test_parser_rust(basic_types, derived_types):

    with DIP() as dip:
        dip.add_string(basic_types)
        dip.add_string(derived_types)
        env = dip.parse()
    with ExportConfigRust(env) as exp:
        assert exp.parse() == """
pub const SIMULATION_NAME: &str = "Configuration test";
pub const SIMULATION_OUTPUT: bool = true;
pub const BOX_HEIGHT: f64 = 15.0;
pub const NUM_CELLS: i32 = 100;
pub const BOX_WIDTH: f32 = 12.0;
pub const DENSITY: f64 = 23.0;
pub const NUM_GROUPS: u64 = 2399495729;
        """.strip()
        
def test_parser_fortran(basic_types, derived_types):

    with DIP() as dip:
        dip.add_string(basic_types)
        dip.add_string(derived_types)
        env = dip.parse()
    with ExportConfigFortran(env) as exp:
        assert exp.parse() == """
module ConfigurationModule
  implicit none

  character(len=20), parameter :: SIMULATION_NAME = "Configuration test";
  logical, parameter :: SIMULATION_OUTPUT = .true.;
  real(kind=8), parameter :: BOX_HEIGHT = 15.0;
  integer, parameter :: NUM_CELLS = 100;
  real, parameter :: BOX_WIDTH = 12.0;
  real(kind=16), parameter :: DENSITY = 23.0;
  integer(kind=8), parameter :: NUM_GROUPS = 2399495729;

end module ConfigurationModule
        """.strip()
        
def test_parser_yaml(basic_types, derived_types, arrays):

    with DIP() as dip:
        dip.add_string(basic_types)
        dip.add_string(derived_types)
        dip.add_string(arrays)
        env = dip.parse()
    with ExportConfigYAML(env) as exp:
        assert exp.parse() == """
box.height:
  unit: cm
  value: 15.0
box.width:
  unit: cm
  value: 12.0
density:
  unit: g/cm3
  value: 23.0
num_cells: 100
num_groups: 2399495729
primes:
- 3
- 5
- 7
simulation.name: Configuration test
simulation.output: true
sizes:
  unit: cm
  value:
  - 23.4
  - 46.0
  - 96.4
        """.strip()
        exp.select("box.width")
        assert exp.parse(units=False) == """
width: 12.0
        """.strip()
        
def test_parser_toml(basic_types, derived_types, arrays):

    with DIP() as dip:
        dip.add_string(basic_types)
        dip.add_string(derived_types)
        dip.add_string(arrays)
        env = dip.parse()
    with ExportConfigTOML(env) as exp:
        assert exp.parse() == """
"simulation.name" = "Configuration test"
"simulation.output" = true
num_cells = 100
num_groups = 2399495729
primes = [ 3, 5, 7,]

["box.height"]
value = 15.0
unit = "cm"

["box.width"]
value = 12.0
unit = "cm"

[density]
value = 23.0
unit = "g/cm3"

[sizes]
value = [ 23.4, 46.0, 96.4,]
unit = "cm"
        """.strip()
        exp.select("box.width")
        assert exp.parse(units=False) == """
width = 12.0
        """.strip()
        
def test_parser_json(basic_types, derived_types, arrays):

    with DIP() as dip:
        dip.add_string(basic_types)
        dip.add_string(derived_types)
        dip.add_string(arrays)
        env = dip.parse()
    with ExportConfigJSON(env) as exp:
        assert exp.parse(indent=2) == """
{
  "simulation.name": "Configuration test",
  "simulation.output": true,
  "box.height": {
    "value": 15.0,
    "unit": "cm"
  },
  "num_cells": 100,
  "box.width": {
    "value": 12.0,
    "unit": "cm"
  },
  "density": {
    "value": 23.0,
    "unit": "g/cm3"
  },
  "num_groups": 2399495729,
  "primes": [
    3,
    5,
    7
  ],
  "sizes": {
    "value": [
      23.4,
      46.0,
      96.4
    ],
    "unit": "cm"
  }
}
        """.strip()
        exp.select("box.width")
        assert exp.parse(units=False) == """
{"width": 12.0}
        """.strip()
        
def test_parser_bash(basic_types, derived_types, arrays):

    with DIP() as dip:
        dip.add_string(basic_types)
        dip.add_string(derived_types)
        dip.add_string(arrays)
        env = dip.parse()
    with ExportConfigBash(env) as exp:
        assert exp.parse() == """
export SIMULATION_NAME="Configuration test"
export SIMULATION_OUTPUT=0
export BOX_HEIGHT=15.0
export NUM_CELLS=100
export BOX_WIDTH=12.0
export DENSITY=23.0
export NUM_GROUPS=2399495729
export PRIMES=("3" "5" "7")
export SIZES=("23.4" "46.0" "96.4")
        """.strip()
        
def test_selection(basic_types):

    # select nodes using a query
    with DIP() as dip:
        dip.add_string(basic_types)
        env = dip.parse()
    with ExportConfigC(env) as exp:
        exp.select(query="box.*")        
        assert exp.parse() == """
#ifndef CONFIG_H
#define CONFIG_H

const double HEIGHT = 15.0;

#endif /* CONFIG_H */
        """.strip()

    # select nodes by tags
    with DIP() as dip:
        dip.add_string(basic_types)
        env = dip.parse()
    with ExportConfigC(env) as exp:
        exp.select(tags=["selection"])        
        assert exp.parse() == """
#ifndef CONFIG_H
#define CONFIG_H

const int NUM_CELLS = 100;

#endif /* CONFIG_H */
        """.strip()

def test_saving(file_pdf, basic_types, derived_types):

    with DIP() as dip:
        dip.add_string(basic_types)
        dip.add_string(derived_types)
        env = dip.parse()
    with ExportConfigC(env) as exp:
        exp.parse() 
        exp.save(file_pdf)
        
        with open(file_pdf,'r') as f:
            assert f.read() == """
#ifndef CONFIG_H
#define CONFIG_H

#include <stdbool.h>

const char* SIMULATION_NAME = "Configuration test";
const bool SIMULATION_OUTPUT = true;
const double BOX_HEIGHT = 15.0;
const int NUM_CELLS = 100;
const float BOX_WIDTH = 12.0;
const long double DENSITY = 23.0;
const unsigned long long int NUM_GROUPS = 2399495729;

#endif /* CONFIG_H */
            """.strip()
            
def test_renaming(basic_types):
    
    with DIP() as dip:
        dip.add_string(basic_types)
        env = dip.parse()
    with ExportConfigC(env, rename=False) as exp:
        assert exp.parse() == """
#ifndef CONFIG_H
#define CONFIG_H

#include <stdbool.h>

const char* simulation.name = "Configuration test";
const bool simulation.output = true;
const double box.height = 15.0;
const int num_cells = 100;

#endif /* CONFIG_H */
        """.strip()